# -*- coding: utf-8 -*-
xhtmlInput = '''
<p style="text-indent: 26px">Considérant qu’il y a lieu de fixer le traitement individuel de <strong><span style="background-color:Yellow">...</span>, né(e) le <span style="background-color:Yellow">...</span></strong> au grade <span style="background-color:Yellow">...</span>, niveau <span style="background-color:Yellow">...</span>, temps <span style="background-color:Yellow">...</span>, dont la carrière administrative est décrite à la fiche individuelle ci-après&nbsp;;</p>

<p>Vu le code de la démocratie locale et de la décentralisation&nbsp;;&nbsp;&nbsp;</p>

<table border="1" cellpadding="1" cellspacing="1" style="height:78px; width:17cm">
    <tbody>
        <tr>
            <td style="text-align:center; vertical-align:middle"><strong>FICHE INDIVIDUELLE DE L’AGENT</strong></td>
        </tr>
        <tr>
            <td style="text-align:center; vertical-align:middle">Né(e) le&nbsp;: <span style="background-color:Yellow">...</span>, Etat civil&nbsp;: <span style="background-color:Yellow">...</span></td>
        </tr>
        <tr>
            <td style="text-align:center; vertical-align:middle"><strong>SERVICES ADMISSIBLES</strong></td>
        </tr>
    </tbody>
</table>

<table border="1" cellpadding="1" cellspacing="1" style="width:17cm">
    <tbody>
        <tr>
            <td colspan="4" rowspan="1" style="background-color:rgb(255, 255, 255)">Supplément pour ancienneté garantie</td>
            <td colspan="5" rowspan="1" style="background-color:rgb(255, 255, 255)">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><u>Nature</u></td>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:bottom"><u>Période (bottom)</u></td>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>DUREE PAR<br />
            GROUPE</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>Coef-<br />
            ficient</strong></td>
            <td colspan="2" style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>VALORISATION<br />
            BAREMIQUE</strong></td>
        </tr>
        <tr>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">Fonction</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">Employeur</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">du</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">au</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>ANS</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>MOIS</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle">&nbsp;</td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>ANS</strong></td>
            <td style="background-color:rgb(204, 204, 204); text-align:center; vertical-align:middle"><strong>MOIS</strong></td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
            <td><strong>TOTAL :</strong></td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" rowspan="1" style="background-color:rgb(204, 204, 204); text-align:right; vertical-align:middle">Ancienneté pécuniaire au <span style="background-color:Yellow">...</span></td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" rowspan="1" style="background-color:rgb(204, 204, 204); text-align:right; vertical-align:middle">Augmentation les <span style="background-color:Yellow">...</span></td>
            <td>&nbsp;</td>
            <td>&nbsp;</td>
        </tr>
    </tbody>
</table>

<table cellspacing="1" style="width:720px">
    <tbody>
        <tr>
            <td colspan="9" style="height:23px; width:720px">FICHE INDIVIDUELLE DE L’AGENT</td>
        </tr>
        <tr>
            <td colspan="9" style="height:23px; width:720px">Né(e) le&nbsp;: ..., Etat civil&nbsp;: ...</td>
        </tr>
        <tr>
            <td colspan="9" style="height:23px; width:720px">SERVICES ADMISSIBLES</td>
        </tr>
        <tr>
            <td colspan="4" style="height:23px; width:320px">Supplément pour ancienneté garantie</td>
            <td colspan="5" style="width:400px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="2" style="height:52px; width:160px">Nature</td>
            <td colspan="2" style="width:160px">Période</td>
            <td colspan="2" style="width:160px">DUREE PAR<br />
            GROUPE</td>
            <td style="width:80px">Coef-<br />
            ficient</td>
            <td colspan="2" style="width:160px">VALORISATION<br />
            BAREMIQUE</td>
        </tr>
        <tr>
            <td style="height:23px; width:80px">Fonction</td>
            <td style="width:80px">Employeur</td>
            <td style="width:80px">du</td>
            <td style="width:80px">au</td>
            <td style="width:80px">ANS</td>
            <td style="width:80px">MOIS</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp; ANS</td>
            <td style="width:80px">MOIS</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td style="height:21px; width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" style="height:23px; width:560px">&nbsp;</td>
            <td style="width:80px">TOTAL :</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" style="height:23px; width:560px">Ancienneté pécuniaire au ...</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
        <tr>
            <td colspan="7" style="height:23px; width:560px">Augmentation les ...</td>
            <td style="width:80px">&nbsp;</td>
            <td style="width:80px">&nbsp;</td>
        </tr>
    </tbody>
</table>

<p><br />
Échelles barémiques&nbsp;: <span style="background-color:Yellow">...</span> – <span style="background-color:Yellow">...</span> – <span style="background-color:Yellow">...</span> (niveau - grade - fonction)</p>

<ul>
    <li>Minimum&nbsp;: <span style="background-color:Yellow">...</span></li>
    <li>Maximum&nbsp;: <span style="background-color:Yellow">...</span></li>
    <li>Développement&nbsp;:&nbsp;
    <ul>
        <li><span style="background-color:Yellow">.. x .. x ..</span></li>
        <li>&nbsp;</li>
    </ul>
    </li>
</ul>
'''



'''Loads and returns a Model instance'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import sys, inspect, pathlib, importlib

from appy.tr import po
from appy.model import Model
from appy.ui.layout import Layouts
from appy.model.fields import Field
from appy.model.fields.ref import Ref
from appy.model.fields.text import Text
from appy.model.fields.phase import Page
from appy.model.workflow import standard
from appy.model.meta.class_ import Class
from appy.model.fields.string import String
from appy.model.workflow.state import State
from appy.model.meta.workflow import Workflow
from appy.model.fields.computed import Computed

# Errors - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
NO_SEARCH = "Note that Appy classes and workflows are not searched within " \
            "your app's __init__.py file, nor within sub-modules."
NO_CLASS  = 'Warning: no Appy class was found in your app @%%s. %s' % NO_SEARCH
DUP_NAME  = '2 or more Appy classes and/or workflows named "%s" have been ' \
            'found in your model. This is not allowed, even if they are in ' \
            'separate Python modules.'
LOADED_RT = 'Model loaded (base+app) - %d classes - %d workflows - %s labels.'
LOADED_MT = 'Model loaded (app) - %d class(es) - %d workflow(s).'
NOT_WF    = '"%s", defined as workflow for class "%s", is not a workflow class.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Loader:
    '''Reads the application model, completes it with the base Appy model, and
       returns a appy.model.Model instance.'''

    def __init__(self, config, appConfig, logger, appOnly):
        # The model config (a appy.model.Config instance)
        self.config = config
        # The root app config
        self.appConfig = appConfig
        # A logger. If None, output will go to stdout.
        self.logger = logger
        # If p_appOnly is True, we will not load the Appy base model: only the
        # classes and workflows defined in the app will be loaded. This
        # minimalist model loading is performed by the appy.tr.updater.Updater
        # (called by <appFolder>/bin/make), for creating or updating the app's
        # i18n files.
        self.appOnly = appOnly

    def log(self, message):
        '''Logs p_message to p_self.logger if it exists, to stdout else'''
        if self.logger:
            self.logger.info(message)
        else:
            print(message)

    def determineMetaClass(self, class_):
        '''If p_class_ is:
           * a Appy class, this method returns meta-class Class;
           * a Appy workflow, this method returns meta-class Wokflow;
           * none of it, this method returns None.
        '''
        # If p_class_ declares at least one static attribute being an instance of
        # appy.fields.Field, it will be considered a Appy class. If it declares
        # at least one static attribute being an instance of
        # appy.fields.workflow.State, it will be considered a Appy workflow.
        for attr in class_.__dict__.values():
            if isinstance(attr, Field): return Class
            elif isinstance(attr, State): return Workflow

    def complete(self, model):
        '''After the model loader has loaded the app's model (m_run), Appy base
           classes (see Model.baseClasses) may have been redefined or not at the
           app level. In order to have a complete model in self.classes, we must
           add all base classes not being redefined at the app level.

           We must also load standard workflows into self.workflows.
        '''
        # Complete self.classes
        for name in Model.baseClasses:
            if name not in model.classes:
                # Import the corresponding base class
                exec('from appy.model.%s import %s' % (name.lower(), name))
                pythonClass = eval(name)
                model.classes[name] = Class(pythonClass, isBase=True)
        # Complete self.workflows
        for name in Model.baseWorkflows:
            model.workflows[name] = Workflow(getattr(standard, name),
                                             isBase=True)

    def updateRefs(self, model):
        '''For "base" and "over" classes, Refs may still point to "base"
           classes. For example, Tool.users may point to appy.model.user.User
           instead of "over" class someApp.user.User. This method updates Refs
           accordingly.'''
        for class_ in model.classes.values():
            # Ignore non-base classes
            if class_.type == 'app': continue
            # Browse forward Ref fields
            for field in class_.fields.values():
                if (field.type != 'Ref') or field.isBack: continue
                # Get the target class
                refClass = model.classes[field.class_.__name__]
                # If this class has been overridden, we must update it with the
                # overridden class.
                if refClass.type == 'over':
                    field.class_ = refClass.python
                    # Update the back ref, too
                    if field.back:
                        field.back.class_ = class_.python

    def linkWorkflows(self, model):
        '''Associate classes and workflows at the model level, based on static
           attributes "workflows" defined on classes, and check that such
           classes are really workflows.'''
        for class_ in model.classes.values():
            workflow = getattr(class_.python, 'workflow', standard.Anonymous)
            name = workflow.__name__
            # Get the corresponding Workflow instance
            workflowClass = model.workflows.get(name)
            if not workflowClass:
                raise Model.Error(NOT_WF % (name, class_.name))
            # Make the link
            class_.workflow = workflowClass

    def setTranslationFields(self, pot, translationClass, message, page):
        '''Sets, on this p_translationClass, 2 fields on p_page:
           (1) a Computed field allowing to display the text in the source
               language;
           (2) an editable String field for storing the corresponding i18n
               p_message.
        '''
        # Create the Computed field displaying the text in the source message
        pythonClass = translationClass.python
        labelField = Computed(method=pythonClass.computeLabel,
                       page=page, show=pythonClass.showField, layouts=Layouts.f)
        name = '%s_label' % message.id
        labelField.init(translationClass, name)
        translationClass.fields[name] = labelField
        # Create the String field for storing the translation
        params = {'page': page, 'layouts': Layouts.f,
                  'show': pythonClass.showField}
        # Scan all messages corresponding to p_messageId from all translation
        # files corresponding to p_pot. We will define field length from the
        # longer found message content.
        maxLine = 100 # We suppose a line is 100 chars long
        width = 0
        height = 0
        potBase = pot.stem
        for fileName, poFile in self.translationFiles.items():
            # Ignore po files not corresponding to p_pot
            if not fileName.startswith('%s-' % potBase): continue
            msgContent = poFile.messages[message.id].msg
            # Compute width
            width = max(width, len(msgContent))
            # Compute height (a "\n" counts for one line)
            mHeight = int(len(msgContent)/maxLine) + msgContent.count('<br/>')
            height = max(height, mHeight)
        if height < 1:
            # Propose a one-line field
            params['width'] = width
            fieldClass = String
        else:
            # Propose a multi-line field, or a very-long-single-lined field
            fieldClass = Text
            params['height'] = height
        field = fieldClass(**params)
        field.init(translationClass, message.id)
        translationClass.fields[message.id] = field

    def updateTranslations(self, model):
        '''Injects, on class Translation, a field for every defined i18n
           label. Returns the total number of i18n labels.'''
        # i18n labels are collected from 3 sources:
        # ----------------------------------------------------------------------
        # Name       | Location   | Description
        # ----------------------------------------------------------------------
        # Appy.pot   | appy/tr/po | The labels used in the base Appy classes.
        # <app>.pot  | <app>/tr   | The labels automatically generated by Appy
        #            |            | from the app model.
        # Custom.pot | <app>/tr   | The custom labels managed "by hand" by the
        #            |            | app's developer.
        # ----------------------------------------------------------------------
        # Translations for Appy.pot labels are not modifiable as-is by the app's
        # user. If the app's developer wants to make some Appy.pot label
        # modifiable, he must define its homonym in Custom.pot.
        # As a preamble, load translations from po(t) files
        trPath = self.config.appPath / 'tr'
        self.translationFiles = po.load(path=trPath, pot=False)
        r = 0
        # Get the Translation class
        translationClass = model.classes['Translation']
        # Read the 3 "pot" files
        appyPot = pathlib.Path(inspect.getfile(po)).parent / 'Appy.pot'
        appPot = trPath / ('%s.pot' % self.config.appPath.stem)
        customPot = trPath / 'Custom.pot'
        for pot in (appyPot, appPot, customPot):
            # Get messages from the current pot file
            messages = po.Parser(pot).parse().messages
            r += len(messages)
            # Manage messages from Appy.pot
            if pot.stem == 'Appy':
                for message in messages.values():
                    # For every message, create a hidden String field. Indeed,
                    # the user will not be allowed to update them (but will be
                    # allowed to modify translations from the app's pot file,
                    # see below).
                    field = String(show=False, layouts='f')
                    field.init(translationClass, message.id)
                    # Add it to class Translation
                    translationClass.fields[message.id] = field
            # Manage messages from app's pot files
            else:
                if pot.stem == 'Custom':
                    phase = 'custom'
                    pageName = 'a'
                else:
                    phase = 'main'
                    pageName = '1'
                page = Page(pageName, phase=phase, sticky=True)
                i = 0
                for message in messages.values():
                    # If this message overrides one from Appy.pot, remove the
                    # overridden field added from Appy.pot. This way, we keep
                    # app's fields in the right order.
                    if message.id in translationClass.fields:
                        del(translationClass.fields[message.id])
                    i += 1
                    # For every message, create 2 fields:
                    # - a Computed used to display the text to translate;
                    # - a String holding the translation in itself.
                    self.setTranslationFields(pot, translationClass,
                                              message, page)
                    if (i % model.config.ui.translationsPerPage) == 0:
                        # A new page must be defined
                        page = Page(page.nextName(), phase=page.phase,
                                    sticky=True)
        return r

    def run(self):
        '''Loads the model'''
        config = self.config
        # Search Appy classes and workflow, in Python files at the root of the
        # App module, __init__.py excepted.
        pythonFiles = []
        for path in config.appPath.glob('*.py'):
            if path.name != '__init__.py':
                pythonFiles.append(path)
        # The app may contain no Python file and only use the base Appy model.
        # Import every found Python file, searching for Appy classes/workflows.
        classType = type(Model)
        classes = {} # ~{s_name: Class}~
        workflows = {} # ~{s_name: Workflow}~
        # Import all modules before reading classes in it. That way, cross-
        # module back references are correctly initialized. Ensure the app's
        # path is in sys.path.
        basePath = str(self.config.appPath.parent)
        if basePath not in sys.path:
            sys.path.insert(0, basePath)
        modules = []
        for path in pythonFiles:
            # We don't use function appy.model.utils.importModule for loading
            # app modules here. Indeed, "importModule" loads a fresh copy of a
            # module, instead of returning the module as already loaded via an
            # "import" statement. But we specifically need to retrieve the
            # already loaded modules, because their classes were already patched
            # with back references.
            name = '%s.%s' % (config.appName, path.stem)
            module = importlib.import_module(name)
            modules.append(module)
        # Browse now the loaded modules
        for module in modules:
            # Find classes within this module
            for element in module.__dict__.values():
                if (type(element) != classType): continue
                # Ignore non-classes or classes imported from other modules
                if (type(element) != classType) or \
                   (element.__module__ != module.__name__): continue
                # Ignore non-Appy classes
                metaClass = self.determineMetaClass(element)
                if not metaClass: continue
                # Add the classes to "classes" or "workflows"
                name = element.__name__
                if (name in classes) or (name in workflows):
                    raise Model.Error(DUP_NAME % name)
                d = classes if metaClass == Class else workflows
                d[name] = metaClass(element, appOnly=self.appOnly)
        # Display a warning if no class was found
        if not classes:
            self.log(NO_CLASS % str(config.appPath))
        # Create the Model instance
        model = Model(self.appConfig, classes, workflows)
        if not self.appOnly:
            # Complete the model with base classes from appy.model
            self.complete(model)
            # Update Refs with "over" classes if defined in the app
            self.updateRefs(model)
            # Link classes with their workflows
            self.linkWorkflows(model)
            # Load translation files and inject every message ID as a field on
            # the Translation class.
            labelsCount = self.updateTranslations(model)
            # Store grantable roles
            model.grantableRoles = model.getRoles(local=False, grantable=True)
            # Inject virtual attributes for every field in every class, in the
            # form of properties.
            for class_ in classes.values(): class_.injectProperties()
            self.log(LOADED_RT % (len(model.classes), len(model.workflows),
                                  labelsCount))
        else:
            self.log(LOADED_MT % (len(model.classes), len(model.workflows)))
        # On server init, also return the translation files, that were already
        # requested while loading the model, in order to determine the length of
        # fields storing translations.
        return model if self.appOnly else (model, self.translationFiles)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import appy
import os.path

# ------------------------------------------------------------------------------
appyPath = os.path.realpath(os.path.dirname(appy.__file__))
od = 'application/vnd.oasis.opendocument'
ms = 'application/vnd.openxmlformats-officedocument'
ms2 = 'application/vnd.ms'

mimeTypes = {'odt': '%s.text' % od,
             'ods': '%s.spreadsheet' % od,
             'doc': 'application/msword',
             'rtf': 'text/rtf',
             'pdf': 'application/pdf'
             }
mimeTypesExts = {
    '%s.text' % od:        'odt',
    '%s.spreadsheet' % od: 'ods',
    'application/msword':  'doc',
    'text/rtf':            'rtf',
    'application/pdf':     'pdf',
    'image/png':           'png',
    'image/jpeg':          'jpg',
    'image/pjpeg':         'jpg',
    'image/gif':           'gif',
    '%s.wordprocessingml.document' % ms: 'docx',
    '%s.spreadsheetml.sheet' % ms: 'xlsx',
    '%s.presentationml.presentation' % ms: 'pptx',
    '%s-excel' % ms2:      'xls',
    '%s-powerpoint' % ms2: 'ppt',
    '%s-word.document.macroEnabled.12' % ms2: 'docm',
    '%s-excel.sheet.macroEnabled.12' % ms2: 'xlsm',
    '%s-powerpoint.presentation.macroEnabled.12' % ms2: 'pptm'
}

# ------------------------------------------------------------------------------
class UnmarshalledFile:
    '''Used for producing file objects from a marshalled Python object.'''
    def __init__(self):
        self.name = '' # The name of the file on disk
        self.mimeType = None # The MIME type of the file
        self.content = '' # The binary content of the file or a file object
        self.size = 0 # The length of the file in bytes.

class UnicodeBuffer:
    '''With StringIO class, I have tons of encoding problems. So I define a
       similar class here, that uses an internal unicode buffer.'''
    def __init__(self):
        self.buffer = []
    def write(self, s):
        if s == None: return
        if isinstance(s, unicode):
            self.buffer.append(s)
        elif isinstance(s, str):
            self.buffer.append(s.decode('utf-8'))
        else:
            self.buffer.append(unicode(s))
    def getValue(self):
        return u''.join(self.buffer)
# ------------------------------------------------------------------------------



# -*- coding: utf-8 -*-
xhtmlInput = '''<table>
 <tr>
  <td>Default cell style</td>
  <td><p>This should also get the default cell style</p></td>
 </tr>
 <tr>
    <td>
      <p>First line</p>
      <p>Second line</p>
    </td>
    <td>
      <p class="Signature">Must get style Signature</p>
    </td>
  </tr>
  <tr>
    <td><p>Para1</p><span>Para2</span><p>Para3</p></td>
    <td><p>Para1</p><p>Para2</p><p>Para3</p>Para4</td>
  </tr>
</table>
'''



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import time, os, os.path
from DateTime import DateTime

import appy
from appy.px import Px
from appy.utils import path as putils
from appy.pod.renderer import Renderer
from appy.model.searches import Search
from appy.utils import string as sutils
from appy.utils.string import Normalize
from appy.model.utils import Object as O
from appy.model.fields import Field, Show
from appy.ui.layout import Layouts, Layout
from appy.model.fields.file import FileInfo
from appy.pod import PodError, styles_manager, formatsByTemplate

# Messages - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
POD_ERROR = 'An error occurred while generating the document. Please contact ' \
            'the system administrator.'
NO_TPL    = 'Please specify a pod template in field "template".'
UNAUTH    = 'You are not allowed to perform this action.'
TPL_INEX  = 'Template not found at %s.'
FREEZ_ERR = 'Error while trying to freeze a "%s" file in pod field "%s" (%s).'
FREEZ_FER = 'Server error. Please contact the administrator.'
RENDERING = "Rendering pod %s :: %s (in %s)..."
RENDERED  = "%s:%s (%s) rendered in %.2fsec."
MAIL_NO_R = 'Mailing "%s" contains no recipient.'
MAIL_NO_U = 'Mailing "%s": inexistent user or no email for "%s".'
MAIL_NO_F = 'Mailing "%s" contains no recipient (after removing wrong ' \
            'entries, see above).'
OVERWRIT  = '%s: overwriting %s...'
FREEZ_ODT = 'Freezing the ODT version...'
FREEZ_OV  = 'Freeze: overwriting %s...'
FROZEN_OK = 'Frozen at %s.'
UNFROZ_OK = 'Removed (unfrozen) %s.'
UPLOAD_OK = 'Uploaded at %s.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''POD-specific configuration'''
    def __init__(self):
        # We suppose the Python interpreter running the Appy has library UNO
        # allowing to contact LibreOffice in server mode. If it is not the case,
        # specify here the absolute path to a UNO-compliant Python interpreter.
        self.unoEnabledPython = None
        # The port for contacting LibreOffice
        self.libreOfficePort = 2002

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Mailing:
    '''Represents a mailing list as can be used by a pod field (see below)'''
    def __init__(self, id, name=None, logins=None, subject=None, body=None):
        # The mailing list ID, an element among available mailings as defined in
        # Pod.mailing.
        self.id = id
        # The mailing list name, as shown in the user interface
        self.name = name
        # The list of logins that will be used as recipients for sending
        # emails.
        self.logins = logins
        # The mail subject
        self.subject = subject
        # The mail body
        self.body = body

    def complete(self, field, o, template, mailing, mailText, forUi):
        '''Complete the mailing information'''
        _ = o.translate
        # Guess the name from the ID if no name is defined
        self.name = self.name or sutils.produceNiceMessage(self.id)
        mapping = None
        # Compute the mail subject when absent
        if not self.subject:
            # Give a predefined subject
            mapping = {'site': o.siteUrl,
                       'title':  o.getValue('title', formatted='shown'),
                       'template': field.getTemplateName(o, template)}
            self.subject = _('podmail_subject', mapping=mapping)
        if forUi:
            self.subject = '<b>%s</b>: %s</br/>' % \
                           (_('email_subject'), self.subject)
        # Compute the mail body when absent
        self.body = self.body or mailText
        if not self.body:
            # Give a predefined body
            if not mapping: mapping = {'site': o.siteUrl}
            fmt = 'html' if forUi else 'text'
            self.body = _('podmail_body', mapping=mapping, format=fmt)

    def getConfirmLabel(self, _):
        '''We misuse the "confirm" popup to allow the user to modify the mail
           body in its "comment" field. We misuse this field's label to
           integrate the mail subject in it (so the user can see it). This
           method returns this "label".'''
        return '%s<br/><b>%s</b>' % (self.subject, _('email_body'))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class ImageFinder:
    '''Used by the POD renderer (XHTML conversion) to retrieve paths to local
       images instead of performing HTTP GETs to the server running POD.'''

    def __init__(self, o):
        self.o = o
        self.siteUrl = o.siteUrl

    def find(self, url):
        '''If p_url is local to the site running this code, this method returns
           the absolute path to the file on disk. Else, it returns None, letting
           POD perform a HTTP GET request to the "external" web server.'''
        # Return None if the image is not local to this webserver
        base = self.siteUrl
        if not url.startswith(base): return
        # Get the image path on disk. We make the assumption that the file is
        # stored in a File field on some object, so URL is supposed to be of the
        # form
        #               <base>/<objectId>/<fieldName>/download
        o = info = None
        for part in url[len(base):].split('/'):
            # Ignore empty parts
            if not part: continue
            # Get the object on which the File field is defined
            if o is None:
                o = self.o.getObject(part)
                if o is None: return
                continue
            # Get the FileInfo instance on this object
            if info is None:
                info = getattr(o, part)
                if not info: return
                break
        # If we are here, we can determine the path to the file on disk
        return str(info.getFilePath(o))

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Pod(Field):
    '''A pod is a field allowing to produce a (PDF, ODT, Word, RTF...) document
       from data contained in Appy class and linked objects or anything you
       want to put in it. It is the way gen uses pod.'''

    # Some methods wil be traversable
    traverse = Field.traverse.copy()

    # Make some useful classes accessible here
    TableProperties = styles_manager.TableProperties
    BulletedProperties = styles_manager.BulletedProperties
    NumberedProperties = styles_manager.NumberedProperties

    # Default Pod config
    defaultConfig = Config()

    # Default output formats for POD templates, by template type
    allFormats = formatsByTemplate

    # Getting a pod value is something special: disable the standard Appy
    # machinery for this.
    customGetValue = True

    class Layouts(Layouts):
        '''Pod-specific layouts'''
        # Right-aligned layouts, convenient for pod fields exporting search
        # results or multi-template pod fields.
        r = Layouts(view=Layout('lf!', css_class='podTable')) # "r"ight
        rf= Layouts(view=Layout('f!', css_class='podTable')) # "r"ight, no label
        l = Layouts(view= Layout('lf;', css_class='podTable')) # "l"eft
        # "r"ight "m"ulti-template (where the global field label is not used)
        rm = Layouts(view=Layout('f!', css_class='podTable'))
        # Inline layout
        inline = Layouts(view=Layout('f', width=None, css_class='inline',
                         align=''))

    # Icon allowing to generate a given template in a given format
    pxIcon = Px('''
     <a var="fmtSuffix='●' if frozen else '';
             linkId=field.getLinkId(confirm, id, info, fmt);
             linkGetter='getNode(%s)' % q(linkId) if confirm else 'this';
             js='generatePod(%s,%s,%s,%s,%s,%s,null,%s)' % \
                 (linkGetter, q(o.url), q(name), q(info.template), q(fmt), \
                 q(tool.Search.encodeForReplay(req, layout)), gc)"
        onclick=":confirm and 'askConfirm(%s,%s,%s)' % (q('script'), \
                 q(js,False), field.labelId) or js"
        title=":field.getIconTitle(o, fmt, frozen)" class="clickable"
        style="position: relative" id=":linkId">
      <img src=":url('download.svg')" class="iconS"/>
      <span class="fmt">:fmt.upper() + fmtSuffix</span>
     </a>''')

    view = cell = Px('''
     <script if="field.confirm">::field.getJsConfirmVar(o)</script>
     <x var="id=o.iid;
             gc=field.getCheckHook(o, req);
             onCell=layout == 'cell'"
        for="info in field.getVisibleTemplates(o)"
        var2="mailings=field.getVisibleMailings(o, info.template);
              lineBreak=((loop.info.nb + 1) % field.maxPerRow) == 0;
              confirm=field.getConfirm(o, info.template)">
      <!-- Show the specific template name when relevant -->
      <span if="not onCell and field.useTemplateName"
            class="pod smaller">:field.getTemplateName(o, info.template)</span>
      <x for="fmt in info.formats"
         var2="freezeAllowed=(fmt in info.freezeFormats) and \
                             (field.show != 'result');
               hasMailings=mailings and (fmt in mailings);
               dropdownEnabled=freezeAllowed or hasMailings;
               frozen=fmt in info.frozenFormats">
       <!-- A clickable icon if no freeze action is allowed and no mailing is
            available for this format -->
       <x if="not dropdownEnabled">:field.pxIcon</x>
       <!-- A clickable icon and a dropdown menu else -->
       <span if="dropdownEnabled" class="dropdownMenu"
             onmouseover="toggleDropdown(this)"
             onmouseout="toggleDropdown(this,'none')">
        <x>:field.pxIcon</x>
        <!-- The dropdown menu containing freeze actions -->
        <table class="dropdown" width="110px">
         <!-- Unfreeze -->
         <tr if="freezeAllowed and frozen" valign="top">
          <td width="95px">
           <a onclick=":'freezePod(%s,%s,%s,%s,%s)' % (q(o.url), q(name), \
                        q(info.template), q(fmt), q('unfreeze'))"
              class="smaller">:_('unfreezeField')</a>
          </td>
          <td width="15px"><img src=":url('unfreeze')"/></td>
         </tr>
         <!-- (Re-)freeze -->
         <tr if="freezeAllowed" valign="top">
          <td width="85px">
           <a onclick=":'freezePod(%s,%s,%s,%s,%s)' % (q(o.url), q(name), \
                        q(info.template), q(fmt), q('freeze'))"
              class="smaller">:_('freezeField')</a>
          </td>
          <td width="15px"><img src=":url('freeze')"/></td>
         </tr>
         <!-- (Re-)upload -->
         <tr if="freezeAllowed" valign="top">
          <td width="85px">
           <a onclick=":'uploadPod(%s,%s,%s,%s)' % (q(o.url), q(name), \
                                   q(info.template), q(fmt))"
              class="smaller">:_('uploadField')</a>
          </td>
          <td width="15px"><img src=":url('upload')"/></td>
         </tr>
         <!-- Mailing lists -->
         <x if="hasMailings" var2="sendLabel=_('email_send')">
          <tr for="mailing in mailings[fmt]" valign="top"
              var2="mInfo=field.getMailingInfo(o, info.template, mailing,
                                               None, forUi=True)">
           <td colspan="2">
            <a var="js='generatePod(this,%s,%s,%s,%s,%s,null,%s,%s,comment)' % \
                        (q(id), q(name), q(info.template), q(fmt), \
                         q(tool.Search.encodeForReplay(req, layout)), gc, \
                         q(mailing))"
               onclick=":'askConfirm(%s,%s,null,true,400,%s,%s,6)' % \
                          (q('script'), q(js,False), \
                           q(mInfo.getConfirmLabel(_)), q(mInfo.body))"
               title=":sendLabel">
             <img src=":url('email')" align="left" style="margin-right: 2px"/>
             <x>:mInfo.name</x></a>
            </td>
          </tr>
         </x>
        </table>
       </span>
      </x>
      <br if="lineBreak"/>
     </x>''',

     css='''.phase .fmt { position: absolute; left: 27%; bottom: -80%;
                          font-size: 85% }
            .list .fmt { font-weight: normal; margin-right: -8px }
            @keyframes movedown {
              from { top: 0px  }
              50% { top: -5px }
              to { top: 0px }
            }
            .waiting { animation: movedown 0.7s ease infinite }''',

     js='''
      podDownloadStatus = function(node, data) {
        // Checks the status of cookie "podDownload"
        var status = readCookie('podDownload');
        if (status == 'False') return;
        // The download is complete. Stop the timeout.
        clearInterval(podTimeout);
        for (var key in data) node.setAttribute(key, data[key]);
      }

      // Manage query data, if the POD to produce is tied to a search
      manageQueryData = function(f, queryData, checkHook) {
        if (queryData) {
          /* If "queryData" specifies a custom search, get criteria from the
             browser's session storage. */
          var elems = queryData.split(':');
          if (elems[1] == 'customSearch') {
            f.criteria.value = sessionStorage.getItem(elems[0]);
          }
        }
        else if (checkHook) {
          /* If this hook refers to a Search within a Ref field, get search
             params and build query data from it. */
          var ssel = document.getElementById(checkHook + '_ssel');
          if (ssel && ssel.value) {
            // Get the search params corresponding to this search
            var p = document.getElementById('searchResults')['ajax'].params,
                data=[p['className'], p['search'], p['sortKey'], p['sortOrder'],
                      stringFromDict(p['filters'])].join(':');
            f.queryData.value = data;
          }
        }
      }

      // Generate a document from a POD template
      generatePod = function(node, url, fieldName, template, podFormat,
                       queryData, customParams, checkHook, mailing, mailText) {
        var f = document.getElementById('podForm');
        f.action = url + '/' + fieldName + '/generate';
        f.template.value = template;
        f.podFormat.value = podFormat;
        f.queryData.value = queryData;
        f.customParams.value = customParams || '';
        manageQueryData(f, queryData, checkHook);
        if (mailing) {
          f.mailing.value = mailing;
          if (mailText) f.mailText.value = mailText;
        }
        // Transmit value of cookie "showSubTitles"
        f.showSubTitles.value = readCookie('showSubTitles') || 'True';
        // Initialise the status of checkboxes
        setChecked(f, checkHook);
        // Disable the link to prevent double-clicks
        var data = {'class': node.className,
                    'onclick': node.attributes.onclick.value};
        node.setAttribute('onclick', '');
        node.className = 'waiting';
        // Set the pod download cookie. "False" means: not downloaded yet.
        createCookie('podDownload', 'False');
        // Set a timer that will check the cookie value
        podTimeout = window.setInterval(function(){
          podDownloadStatus(node, data)}, 700);
        f.submit();
      }

      // (Un-)freeze a document from a pod template
      freezePod = function(url, fieldName, template, podFormat, action) {
        var f = document.getElementById('podForm');
        f.action = url + '/' + fieldName + '/onFreeze';
        f.template.value = template;
        f.podFormat.value = podFormat;
        f.freezeAction.value = action;
        askConfirm('form', 'podForm', action_confirm);
      }

      // Upload a file for freezing it in a pod field
      uploadPod = function(url, fieldName, template, podFormat) {
        var f = document.getElementById('uploadForm');
        f.action = url + '/' + fieldName + '/upload';
        f.template.value = template;
        f.podFormat.value = podFormat;
        f.uploadedFile.value = null;
        openPopup('uploadPopup');
      }''')

    edit = search = ''

    def __init__(self, validator=None, show=Show.TR, page='main', group=None,
      layouts=None, move=0, readPermission='read', writePermission='write',
      width=None, height=None, maxChars=None, colspan=1, master=None,
      masterValue=None, focus=False, historized=False, mapping=None,
      generateLabel=None, label=None, template=None, templateName=None,
      useTemplateName=None, showTemplate=None, freezeTemplate=None,
      showFrozenOnly=True, maxPerRow=5, context=None, stylesMapping={},
      formats=None, getChecked=None, mailing=None, mailingName=None,
      showMailing=None, mailingInfo=None, view=None, cell=None, edit=None,
      xml=None, translations=None, downloadName=None,
      downloadDisposition='attachment', forceOoCall=False,
      optimalColumnWidths=False, distributeColumns=None, script=None,
      pdfOptions='ExportNotes=True', tabbedCR=False, fonts=None, confirm=False,
      raiseOnError=False, action=None, beforeAction=None):
        # Param "template" stores the path to the pod template(s). If there is
        # a single template, a string is expected. Else, a list or tuple of
        # strings is expected. Every such path must be relative to your
        # application. A pod template name Test.odt that is stored at the root
        # of your app will be referred as "Test.odt" in self.template. If it is
        # stored within sub-folder "pod", it will be referred as "pod/Test.odt".
        if not template: raise Exception(NO_TPL)
        if isinstance(template, str):
            self.template = [template]
        elif isinstance(template, tuple):
            self.template = list(template)
        else:
            self.template = template
        # Param "templateName", if specified, is a method that will be called
        # with the current template (from self.template) as single arg and must
        # return the name of this template. If self.template stores a single
        # template, you have no need to use param "templateName" (excepted if
        # you want to render the POD with a name, in the layout "buttons").
        # In most cases, if you have a single template, simply use the field
        # label to name the template. If you have a multi-pod field (with
        # several templates specified as a list or tuple in param "template"),
        # you will probably choose to hide the field label and use param
        # "templateName" to give a specific name to every template. If
        # "template" contains several templates and "templateName" is None, Appy
        # will produce names from template filenames.
        self.templateName = templateName
        # If you want to use "templateName" hereabove even if self.template
        # contains a single template, set "useTemplateName" to True.
        if useTemplateName is None:
            self.useTemplateName = len(self.template) > 1
        else:
            self.useTemplateName = useTemplateName
        # "showTemplate" determines if the current user may generate documents
        # based on this pod field. More precisely, "showTemplate", if specified,
        # must be a method that will be called with the current template as
        # single arg (one among self.template) and that must return the list or
        # tuple of formats that the current user may use as output formats for
        # generating a document. If the current user is not allowed at all to
        # generate documents based on the current template, "showTemplate" must
        # return an empty tuple/list. If "showTemplate" is not specified, the
        # user will be able to generate documents based on the current template,
        # in any format from self.formats (see below).
        # "showTemplate" comes in addition to self.show. self.show dictates the
        # visibility of the whole field (ie, all templates from self.template)
        # while "showTemplate" dictates the visiblity of a specific template
        # within self.template.
        self.showTemplate = showTemplate
        # "freezeTemplate" determines if the current user may freeze documents
        # normally generated dynamically from this pod field. More precisely,
        # "freezeTemplate", if specified, must be a method that will be called
        # with the current template as single arg and must return the (possibly
        # empty) list or tuple of formats the current user may freeze. The
        # "freezing-related actions" that are granted by "freezeTemplate" are
        # the following. When no document is frozen yet for a given
        # template/format, the user may:
        # - freeze the document: pod will be called to produce a document from
        #   the current database content and will store it in the database.
        #   Subsequent user requests for this pod field will return the frozen
        #   doc instead of generating on-the-fly documents;
        # - upload a document: the user will be able to upload a document that
        #   will be stored in the database. Subsequent user requests for this
        #   pod field will return this doc instead of generating on-the-fly
        #   documents.
        # When a document is already frozen or uploaded for a given
        # template/format, the user may:
        # - unfreeze the document: the frozen or uploaded document will be
        #   deleted from the database and subsequent user requests for the pod
        #   field will again generate on-the-fly documents;
        # - re-freeze the document: the frozen or uploaded document will be
        #   deleted, a new document will be generated from the current database
        #   content and will be frozen as a replacement to the deleted one;
        # - upload a document: the frozen or uploaded document will be replaced
        #   by a new document uploaded by the current user.
        self.freezeTemplate = freezeTemplate
        # If "showFrozenOnly" is True, only formats for which a frozen document
        # exists will be available to the user. This mechanism is used to
        # prevent data from being frozen in one format and being available for
        # real-time computation in another format, which could be incoherent.
        # Note that is has sense (and thus applies) only when there is at least
        # one frozen document. When "showFrozenOnly" is applied, and when there
        # is at least one frozen document, it overrides the list of available
        # formats (as determined by self.showTemplate or self.formats).
        self.showFrozenOnly = showFrozenOnly
        # If p_template contains more than 1 template, "maxPerRow" tells how
        # much templates must appear side by side.
        self.maxPerRow = maxPerRow
        # The context is a dict containing a specific pod context, or a method
        # that returns such a dict.
        self.context = context
        # A global styles mapping that would apply to the whole template(s)
        self.stylesMapping = stylesMapping
        # What are the output formats when generating documents from this pod ?
        self.formats = formats or self.getAllFormats(self.template[0])
        # Parameter "getChecked" can specify the name of a Ref field belonging
        # to the same Appy class, or the term "search". If getChecked is...
        # ----------------------------------------------------------------------
        # the name of | The context of the pod template will contain an
        # a Ref field | additional object, named "_checked". On this object, an
        #             | attribute will be set, whose name is the name of the Ref
        #             | field, and whose value will be the list of the objects
        #             | linked via the Ref field that are currently selected in
        #             | the user interface.
        # ----------------------------------------------------------------------
        #   "search"  | The pod context variable "objects", normally containing
        #             | all search results, will only contain objects being
        #             | checked in the user interface.
        # ----------------------------------------------------------------------
        self.getChecked = getChecked
        # Mailing lists can be defined for this pod field. For every visible
        # mailing list, a menu item will be available in the user interface and
        # will allow to send the pod result as attachment to the mailing list
        # recipients. Attribute p_mailing stores a mailing list's id
        # (as a string) or a list of ids.
        self.mailing = mailing
        if isinstance(mailing, str):
            self.mailing = [mailing]
        elif isinstance(mailing, tuple):
            self.mailing = list(mailing)
        # "showMailing" below determines when the mailing list(s) must be shown.
        # It may store a method accepting a mailing list's id (among
        # self.mailing) and a template (among self.template) and returning the
        # list or tuple of formats for which the pod result can be sent to the
        # mailing list. If no such method is defined, the mailing list will be
        # available for all visible templates and formats.
        self.showMailing = showMailing
        # When it it time to send an email (or preview it), "mailingInfo" gives
        # all the necessary information for this email: recipients, subject
        # body. It must be a method accepting 2 args:
        # - "mailing" is the mailing ID (from self.mailing);
        # - "forUi" is a boolean indicating if we must compute mailing info for
        #           displaying it in the user interface (forUi=True) or for
        #           sending the mail (forUi=False). It is important because, if
        #           True (False), you must translate i18n labels with parameter
        #           format="html" (format="text").
        # Displaying info in the user interface allows the user to preview the
        # mail and modify the predefined mail body if necessary. The method must
        # return a Mailing instance (see class hereabove). Note that the
        # returned Mailing instance can be subsequently modified by the
        # framework.
        self.mailingInfo = mailingInfo
        # "downloadName", if specified, is a method that will be called with
        # the current template (from self.template) as single arg and must
        # return the name of the file as the user will get it once he will
        # download the pod result from its browser. This is for people that do
        # not like the default download name. Do not specify any extension: it
        # will be appended automatically. For example, if your method returns
        # "PodResultForSomeObject", and the pod result is a pdf file, the file
        # will be named "PodResultForSomeObject.pdf". If you specify such a
        # method, you have the responsibility to produce a valid,
        # any-OS-and-any-browser-proof file name. For inspiration, see the
        # default m_getDownloadName method hereafter. If you have several
        # templates in self.template, for some of them where you are satisfied
        # with the default download name, return None.
        self.downloadName = downloadName
        # The field below allow to determine the "disposition" when downloading
        # a pod result. "attachment" by default, it can be set to "inline". But
        # with disposition=inline, Google Chrome and IE may launch a PDF viewer
        # that triggers one or many additional crashing HTTP GET requests.
        self.downloadDisposition = downloadDisposition
        # Normally, when you generate a pod result that is in ODT/ODS format,
        # LibreOffice is not called. But if you want it to be called
        # nevertheless, for example to ensure that all the indexes are
        # up-to-date (including the table of contents), set "forceOoCall" to
        # True. When generating pod results in other formats (pdf, doc, xls...),
        # LibreOffice is always called and indexes are always refreshed.
        self.forceOoCall = forceOoCall
        # For these parameters, see appy.pod.renderer.Renderer's docstring
        self.optimalColumnWidths = optimalColumnWidths
        self.distributeColumns = distributeColumns
        # "script" can be used to customize the process of rendering the POD
        # result via LibreOffice UNO. See option "-s" of appy/pod/converter.py
        # for more information. "script" can hold the absolute path to a Python
        # file or a method that returns it. If you specify a method, it will
        # accept 2 parameters:
        # * template     the name of the current template (important when
        #                multiple templates are in use);
        # * context      the context given to the pod template.
        self.script = script
        # When the result is PDF, options for configuring the production of the
        # PDF file can be given in this attribute (ie, for adding a watermark).
        # More information about available options on class
        # appy.pod.renderer.Renderer, which has an homonym attribute.
        # "pdfOptions" can hold a method as well.
        self.pdfOptions = pdfOptions
        # "tabbedCR" determines how the "carriage return" char is rendered on
        # POD results. See appy.pod.renderer.py for more information.
        self.tabbedCR = tabbedCR
        # The name of the font passed in "fonts" will be injected in all styles
        # from the POD template. It must return a string or be a function that
        # will accept the current template (one among self.template) as unique
        # arg and return the name of the font to inject in it.
        self.fonts = fonts
        # If "confirm" is True, a popup will be shown before generating the
        # pod result. "confirm" may also hold a method, that must accept, as
        # single arg, the name of the current template (indeed, self.template
        # may hold several templates), and must return a boolean value.
        self.confirm = confirm
        # If "raiseOnError" is False (the default), no error is raised:
        # traceback(s) is (are) dumped into the pod result within note(s). If
        # you prefer a real exception to be raised, set this parameter to True.
        # This could be important if, for example, an action is tied to this pod
        # field: you want it to be completely executed or rolled back.
        self.raiseOnError = raiseOnError
        # If you want some action to be executed after the pod result has been
        # generated, set a method in parameter "action". This must be a method
        # accepting parameters:
        # * template     the name of the current template (important when
        #                multiple templates are in use);
        # * context      the context given to the pod template;
        # * format       the output format.
        self.action = action
        # If you want some action to be executed just before the pod result is
        # generated, set a method in parameter "beforeAction". This method's
        # signature must be the same as for parameter "action" hereabove.
        self.beforeAction = beforeAction
        # Call the base constructor
        Field.__init__(self, None, (0,1), None, None, show, page, group,
          layouts, move, False, True, None, None, False, None, readPermission,
          writePermission, width, height, None, colspan, master, masterValue,
          focus, historized, mapping, generateLabel, label, None, None, None,
          None, False, False, view, cell, edit, xml, translations)
        # Param "persist" is False, but actual persistence for this field is
        # determined by freezing.
        self.validable = False

    def getExtension(self, template):
        '''Gets a p_template's extension (".odt" or ".ods"). Because a template
           can simply be a pointer to another template (ie, "Item.odt.variant"),
           the logic for getting the extension is a bit more tricky.'''
        elems = os.path.splitext(template)
        if elems[1] in Pod.allFormats: return elems[1]
        # p_template must be a pointer to another template and has one more
        # extension.
        return os.path.splitext(elems[0])[1]

    def getAllFormats(self, template):
        '''Gets all the output formats that are available for a given
           p_template.'''
        return Pod.allFormats[self.getExtension(template)]

    def isRenderable(self, layout):
        '''A single-template POD is renderable on layout "buttons"'''
        return True if layout != 'buttons' else len(self.template) == 1

    def setTemplateFolder(self, folder):
        '''This methods adds a prefix to every template name in self.template'''
        # This can be useful if a plug-in module needs to replace an application
        # template by its own templates. Here is an example: imagine a base
        # application has a pod field with:

        # self.templates = ["Item.odt", "Decision.odt"]

        # The plug-in module, named "PlugInApp", wants to replace it with its
        # own templates Item.odt, Decision.odt and Other.odt, stored in its sub-
        # folder "pod". Suppose the base pod field is in <podField>. The plug-in
        # will write:

        # <podField>.templates = ["Item.odt", "Decision.odt", "Other.odt"]
        # <podField>.setTemplateFolder('../PlugInApp/pod')

        # The following code is equivalent, will work, but is precisely the kind
        # of things we want to avoid.

        # <podField>.templates = ["../PlugInApp/pod/Item.odt",
        #                           "../PlugInApp/pod/Decision.odt",
        #                           "../PlugInApp/pod/Other.odt"]
        for i in range(len(self.template)):
            self.template[i] = os.path.join(folder, self.template[i])

    def getTemplateName(self, o, fileName):
        '''Gets the name of a template given its p_fileName'''
        r = None
        if self.templateName:
            # Use the method specified in self.templateName
            r = self.templateName(o, fileName)
        # Else, deduce a nice name from p_fileName
        if not r:
            name = os.path.splitext(os.path.basename(fileName))[0]
            r = sutils.produceNiceMessage(name)
        return r

    def getTemplatePath(self, diskFolder, template):
        '''Return the absolute path to some pod p_template, by prefixing it with
           the application path. p_template can be a pointer to another
           template.'''
        # Compute the base path
        if template.startswith('/'):
            # An absolute path: search for it in Appy itself
            baseFolder = os.path.dirname(appy.__file__)
            template = template[1:]
        else:
            baseFolder = diskFolder
        # Compute the absolute path
        r = putils.resolvePath(os.path.join(baseFolder, template))
        if not os.path.isfile(r):
            raise Exception(TPL_INEX % template)
        # Unwrap the path if the file is simply a pointer to another one
        elems = os.path.splitext(r)
        if elems[1] not in Pod.allFormats:
            r = elems[0]
        return r

    def getDownloadName(self, o, template, format=None, queryRelated=False,
                        bypassMethod=False):
        '''Gets the name of the pod result as will be seen by the user that will
           download it. Ensure the returned name is not too long for the OS that
           will store the downloaded file with this name.'''
        # Use method self.downloadName if present and if it returns something
        # for p_template.
        if not bypassMethod and self.downloadName:
            name = self.downloadName(o, template)
            if name: return '%s.%s' % (name, format) if format else name
        # Compute the default download name
        fileName = Normalize.fileName(self.getTemplateName(o, template))[:40]
        if not queryRelated:
            # This is a POD for a single object: personalize the file name with
            # the object title.
            title = o.getShownValue('title')
            fileName = '%s-%s' % (Normalize.fileName(title)[:40], fileName)
        return '%s.%s' % (fileName, format) if format else fileName

    def getVisibleTemplates(self, o):
        '''Returns, among self.template, the template(s) that can be shown'''
        r = []
        if not self.showTemplate:
            # Show them all in the formats specified in self.formats
            for template in self.template:
                r.append(O(template=template, formats=self.formats,
                           freezeFormats=self.getFreezeFormats(o, template)))
        else:
            for template in self.template:
                formats = self.showTemplate(o, template)
                if not formats: continue
                elif isinstance(formats, bool):
                    formats = self.getAllFormats(template)
                elif isinstance(formats, str):
                    formats = (formats,)
                r.append(O(template=template, formats=formats,
                           freezeFormats=self.getFreezeFormats(o, template)))
        # Compute the already frozen documents, and update the available formats
        # accordingly when self.showFrozenOnly is True.
        for info in r:
            frozenFormats = []
            for fmt in info.formats:
                if self.isFrozen(o, info.template, fmt):
                    frozenFormats.append(fmt)
            info.frozenFormats = frozenFormats
            # Replace formats with frozenFormats when relevant
            if frozenFormats and self.showFrozenOnly:
                info.formats = frozenFormats
        return r

    def getVisibleMailings(self, o, template):
        '''Gets, among self.mailing, the mailing(s) that can be shown for
           p_template, as a dict ~{s_format:[s_id]}~.'''
        if not self.mailing: return
        r = {}
        for mailing in self.mailing:
            # Is this mailing visible ? In which format(s) ?
            if not self.showMailing:
                # By default, the mailing is available in any format
                formats = True
            else:
                formats = self.showMailing(o, mailing, template)
            if not formats: continue
            if isinstance(formats, bool): formats = self.getAllFormats(template)
            elif isinstance(formats, str): formats = (formats,)
            # Add this mailing to the result
            for fmt in formats:
                if fmt in r: r[fmt].append(mailing)
                else: r[fmt] = [mailing]
        return r

    def getConfirm(self, o, template):
        '''When producing a result from p_template for p_obj, must we ask a
           confirmation to the user ?'''
        confirm = self.confirm
        return confirm(o, template) if callable(confirm) else confirm

    def getCheckHook(self, o, req):
        '''If p_self.getChecked is in use, return the ID of the DOM node
           containing the list of (un)checked objects.'''
        gc = self.getChecked
        if not gc:
            r = 'null'
        elif gc == 'search':
            r = req.search
            r = '"%s"' % r if r else 'null'
        else:
            if o.allows(o.getField(gc).readPermission):
                r = '"%d_%s"' % (o.iid, gc)
            else:
                r = 'null'
        return r

    def getLinkId(self, confirm, objectId, info, fmt):
        '''Returns an identifier for a link allowing to trigger a pod. This
           identifier is long and is not needed when p_confirm is False.'''
        if not confirm: return ''
        return '%s_%s_%s_%s' % (objectId, self.name,
                                Normalize.fileName(info.template), fmt)

    def getMailingInfo(self, o, template, mailing, mailText, forUi=False):
        '''Gets the necessary information for sending an email to p_mailing
           list, or for getting this info for displaying it in the user
           interface if p_forUi is True.'''
        r = self.mailingInfo(o, mailing, forUi)
        r.complete(self, o, template, mailing, mailText, forUi)
        return r

    def sendMailing(self, o, template, mailing, mailText, attachment):
        '''Sends the emails for m_mailing'''
        info = self.getMailingInfo(o, template, mailing, mailText)
        if not info.logins:
            o.log(MAIL_NO_R % mailing)
            return 'action_ko'
        # Collect logins corresponding to inexistent users and recipients
        missing = []
        recipients = []
        for login in info.logins:
            user = o.search1('User', secure=False, login=login)
            if not user:
                missing.append(login)
                continue
            else:
                recipient = user.getMailRecipient()
                if not recipient:
                    missing.append(login)
                else:
                    recipients.append(recipient)
        if missing:
            o.log(MAIL_NO_U % (mailing, str(missing)))
        if not recipients:
            o.log(MAIL_NO_F % mailing)
            msg = 'action_ko'
        else:
            o.tool.sendMail(recipients, info.subject, info.body, [attachment])
            msg = 'action_done'
        return msg

    def getValue(self, o, name=None, layout=None, template=None, format=None,
                 result=None, queryData=None, computeCustomContext=None,
                 secure=True, executeAction=True, single=None):
        '''For a pod field, getting its value means computing a pod document or
           returning a frozen one.'''
        # A pod field differs from other field types because there can be
        # several ways to produce the field value (ie: self.template can hold
        # various templates; output file format can be odt, pdf,.... We get
        # those precisions about the way to produce the file, either from
        # params, or from default values.
        # * p_template is the specific template, among self.template, that must
        #   be used as base for generating the document;
        # * p_format is the output format of the resulting document;
        # * p_result, if given, must be the absolute path of the document that
        #   will be computed by pod. If not given, pod will produce a doc in the
        #   OS temp folder;
        # * if the pod document is related to a query, the query parameters
        #   needed to re-trigger the query are given in p_queryData, and, if it
        #   is a custom search, criteria are in the request, at key "criteria";
        # * if p_computeCustomContext is True, this special context will be
        #   computed (see m_setCustomContext) and will override any other value
        #   available in the context, including values from the field-specific
        #   context.
        start = time.time()
        template = template or self.template[0]
        format = format or 'pdf'
        # Security check
        if secure and not queryData:
            if self.showTemplate and not self.showTemplate(o, template):
                raise Exception(UNAUTH)
        # Return the possibly frozen document (not applicable for query-related
        # pods).
        if not queryData:
            frozen = self.isFrozen(o, template, format)
            if frozen:
                fileName = self.getDownloadName(o, template, format)
                return FileInfo(frozen, inDb=False, uploadName=fileName)
        # We must call pod to compute a pod document from "template"
        appPath = o.appPath
        # Get the path to the pod template
        templatePath = self.getTemplatePath(str(appPath), template)
        # Get or compute the specific POD context
        specificContext = self.getAttribute(o, 'context')
        # Compute the name of the result file
        if not result:
            result = '%s/%d_%f.%s' % (putils.getOsTempFolder(), o.iid,
                                      time.time(), format)
        # Define parameters to give to the appy.pod renderer
        req = o.req
        tool = o.tool
        podContext = {'tool': tool, 'user': o.user, 'self': o, 'field': self,
                      'now': DateTime(), '_': o.translate, 'appPath': appPath,
                      'template': template, 'req': req, 'config': tool.config}
        # If the pod document is related to a search, replay it and put the
        # result in the pod context.
        if queryData:
            podContext['objects'] = Search.replay(tool, queryData)
            podContext['queryData'] = queryData.split(':')
            podContext['_ref'] = Search.getRefInfo(tool, nameOnly=False)
        # Add the field-specific context if present
        if specificContext: podContext.update(specificContext)
        # Add the custom context when required
        if computeCustomContext:
            self.setCustomContext(podContext, o, req, queryData)
        # Variable "_checked" can be expected by a template but absent (ie,
        # when generating frozen documents).
        if '_checked' not in podContext: podContext['_checked'] = O()
        # Define a potential global styles mapping
        smap = self.stylesMapping
        smap = self.callMethod(o, smap) if callable(smap) else smap
        # Execute the "before" action when relevant
        if executeAction and self.beforeAction:
            self.beforeAction(o, template, podContext, format)
        # Get the optional script to give to the renderer
        script = self.script
        script = script(o, template, podContext) if callable(script) else script
        # Compute PDF options
        options = self.pdfOptions
        options = options(o, template) if callable(options) else options
        # Compute fonts
        fonts = self.fonts
        fonts = fonts(o, template) if callable(fonts) else fonts
        # Compute the renderer's parameters
        rendererParams = {'template': templatePath, 'context': podContext,
          'result': result, 'stylesMapping': smap, 'overwriteExisting': True,
          'forceOoCall': self.forceOoCall, 'raiseOnError': self.raiseOnError,
          'script': script, 'optimalColumnWidths': self.optimalColumnWidths,
          'distributeColumns': self.distributeColumns, 'pdfOptions': options,
          'tabbedCR': self.tabbedCR, 'fonts': fonts,
          'findImage': ImageFinder(o).find}
        config = o.config.pod or Pod.defaultConfig
        if config.unoEnabledPython:
            rendererParams['pythonWithUnoPath'] = config.unoEnabledPython
        if config.libreOfficePort:
            rendererParams['ooPort'] = config.libreOfficePort
        # Launch the renderer
        o.log(RENDERING % (o.id, template, format))
        try:
            renderer = Renderer(**rendererParams)
            renderer.run()
        except PodError as pe:
            if not os.path.exists(result):
                # In some (most?) cases, when OO returns an error, the result is
                # nevertheless generated.
                o.log(str(pe).strip(), type='error')
                return POD_ERROR
        # Give a friendly name for this file
        fileName = self.getDownloadName(o, template, format, queryData)
        # Execute the tied action when relevant
        if executeAction and self.action:
            self.action(o, template, podContext, format)
        # Log the successfull rendering
        o.log(RENDERED % (o.id, self.name, fileName, time.time()-start))
        # Get a FileInfo instance to manipulate the file on the filesystem
        return FileInfo(result, inDb=False, uploadName=fileName)

    def getBaseName(self, template=None):
        '''Gets the "base name" of p_template (or self.template[0] if not
           given). The base name is the name of the template, without path
           and extension. Moreover, if the template is a pointer to another one
           (ie Item.odt.something), the base name integrates the specific
           extension. In the example, the base name will be "ItemSomething".'''
        template = template or self.template[0]
        elems = os.path.splitext(os.path.basename(template))
        if elems[1] in ('.odt', '.ods'):
            r = elems[0] # Item.odt > Item
        else:
            # Item.odt.something > ItemSomething
            r = os.path.splitext(elems[0])[0] + elems[1][1:].capitalize()
        return r

    def getFreezeName(self, template=None, format='pdf', sep='.'):
        '''Gets the name on disk on the frozen document corresponding to this
           pod field, p_template and p_format.'''
        return '%s_%s%s%s' % (self.name,self.getBaseName(template),sep,format)

    def isFrozen(self, o, template=None, format='pdf'):
        '''Is there a frozen document for thid pod field, on p_o, for p_template
           in p_format? If yes, it returns the absolute path to the frozen
           doc.'''
        template = template or self.template[0]
        folder = o.getFolder()
        fileName = self.getFreezeName(template, format)
        r = folder / fileName
        if r.is_file(): return str(r)

    def freeze(self, o, template=None, format='pdf', secure=False, upload=None,
               freezeOdtOnError=True):
        '''Freezes, on p_o, a document for this pod field, for p_template in
           p_format'''
        # If p_secure is False, the security check, based on
        # self.freezeTemplate, is bypassed. If no p_upload file is specified, we
        # re-compute a pod document on-the-fly and we freeze this document.
        # Else, we store the uploaded file.

        # If p_freezeOdtOnError is True and format is not "odt" (has only sense
        # when no p_upload file is specified), if the freezing fails we try to
        # freeze the odt version, which is more robust because it does not
        # require calling LibreOffice.

        # Security check
        if secure and (format not in self.getFreezeFormats(o, template)):
            raise Exception(UNAUTH)
        # Compute the absolute path where to store the frozen document in the
        # database.
        folder = o.getFolder(create=True)
        fileName = self.getFreezeName(template, format)
        result = folder / fileName
        if result.is_file():
            prefix = 'freeze (upload)' if upload else 'freeze'
            o.log(OVERWRIT % (prefix, result))
            result.unlink()
        if not upload:
            # Generate the document
            doc = self.getValue(o, template=template, format=format,
                                result=str(result))
            if isinstance(doc, str):
                # An error occurred, the document was not generated
                o.log(FREEZ_ERR % (format, self.name, doc), type='error')
                if not freezeOdtOnError or (format == 'odt'):
                    raise Exception(FREEZ_FER)
                o.log(FREEZ_ODT)
                # Freeze the ODT version of the document, which does not require
                # to call LibreOffice: the risk of error is smaller.
                fileName = self.getFreezeName(template, 'odt')
                result = folder / fileName
                if result.is_file():
                    o.log(FREEZ_OV % result)
                doc = self.getValue(o, template=template, format='odt',
                                    result=result)
                if isinstance(doc, str):
                    self.log(FREEZ_ERR % ('odt', self.name, doc),
                             type='error')
                    raise Exception(FREEZ_FER)
                o.log(FROZEN_OK % result)
        else:
            # Store the uploaded file in the database
            result = str(result)
            f = open(result, 'wb')
            doc = FileInfo(result, inDb=False)
            doc.replicateFile(upload.value, f)
            f.close()
            o.log(UPLOAD_OK % result)
        return doc

    def unfreeze(self, o, template=None, format='pdf', secure=False):
        '''Unfreezes, on p_o, the document for this pod field, for p_template
           in p_format.'''
        # Security check
        if secure and (format not in self.getFreezeFormats(o, template)):
            raise Exception(UNAUTH)
        # Compute the absolute path to the frozen doc
        folder = o.getFolder()
        fileName = self.getFreezeName(template, format)
        frozen = folder / fileName
        if frozen.is_file():
            frozen.unlink()
            putils.FolderDeleter.deleteEmpty(str(folder))
            o.log(UNFROZ_OK % str(frozen))

    def getFreezeFormats(self, o, template=None):
        '''What are the formats into which the current user may freeze
           p_template?'''
        # One may have the right to edit the field to freeze anything in it
        if not o.allows(self.writePermission): return ()
        template = template or self.template[0]
        # Users (managers included) can perform freeze actions depending on
        # self.freezeTemplate.
        if not self.freezeTemplate: return ()
        r = self.freezeTemplate(o, template)
        if not r:
            return () # Ensure the result is a tuple
        elif r == True:
            return self.getAllFormats(template)
        else:
            return r

    def getIconTitle(self, o, format, frozen):
        '''Get the title of the format icon'''
        r = o.translate(format)
        if frozen:
            r += ' (%s)' % r.translate('frozen')
        return r

    def setCustomContext(self, context, o, req, queryData):
        '''Before calling pod to compute a result, if specific elements must be
           added to the p_context, compute it here. This request-dependent
           method is not called when computing a pod field for freezing it into
           the database.'''
        # Get potential custom params from the request. Custom params must be
        # coded as a string containing a valid Python dict.
        params = req.customParams
        if params:
            context.update(eval(params))
        # Compute the checked objects when relevant
        gc = self.getChecked
        if not gc: return
        # Manage a Ref field
        if gc == 'search':
            # Manage a search. All search results are in p_context['objects'].
            Search.keepCheckedResults(req, context['objects'])
        else:
            # Manage a Ref field
            if queryData:
                # We are in the presence of results from one of the searches
                # defined in Ref.searches. Search results are already in
                # p_context['objects'].
                Search.keepCheckedResults(req, context['objects'])
                objects = context['objects']
            else:
                # The IDs of the tied objects are specified in the request
                ids, unchecked = Search.getCheckedInfo(req)
                objects = []
                tool = o.tool
                for tied in getattr(o, gc):
                    if unchecked: condition = tied.iid not in ids
                    else:         condition = tied.iid in ids
                    if condition and tied.allows('read'):
                        objects.append(tied)
            context['_checked'] = O()
            setattr(context['_checked'], gc, objects)

    def setRequestValue(self, o):
        '''For a Pod field there is not possibility to carry its value in the
           request.'''

    def resetPodCookie(self, o):
        '''Reset the "pod download" cookie indicating that the pod download as
           ended (successfully or not).'''
        o.resp.setCookie('podDownload', 'True')

    def getJsConfirmVar(self, o):
        '''Gets the Javascript variable definition for storing the specific
           confirmation message to show when self.confirm is not False.'''
        prefix = self.labelId
        return 'var %s = "%s";' % (prefix, o.translate('%s_confirm' % prefix))

    traverse['generate'] = 'perm:read'
    def generate(self, o):
        '''Called from the ui for generating a POD result'''
        req = o.req
        template = req.template
        format = req.podFormat
        mailing = req.mailing
        # Generate a (or get a frozen) document
        r = self.getValue(o, template=template, format=format,
                          queryData=req.queryData, computeCustomContext=True)
        if isinstance(r, str):
            # An error has occurred, and "r" contains the error message
            if not mailing: self.resetPodCookie(o)
            return o.goto(o.referer, message=r)
        # "r" contains a FileInfo instance.
        # Return the result to the ui or send a mail with it as attachment ?
        if not mailing:
            self.resetPodCookie(o)
            disposition = self.downloadDisposition
            r.writeResponse(o.H(), disposition=disposition, cache=False)
        else:
            # Send the email(s)
            msg = self.sendMailing(o, template, mailing, req.mailText, r)
            return o.goto(o.referer, message=o.translate(msg))

    traverse['onFreeze'] = 'perm:write'
    def onFreeze(self, o):
        '''UI request to (un)freeze a file for this POD field'''
        req = o.req
        action = req.freezeAction
        if action == 'freeze':
            self.freeze(o, req.template, req.podFormat, secure=True,
                        freezeOdtOnError=False)
        elif action == 'unfreeze':
            self.unfreeze(o, req.template, req.podFormat, secure=True)
        return o.goto(o.referer, message=o.translate('action_done'))

    traverse['upload'] = 'perm:write'
    def upload(self, o):
        '''UI request to upload a file for (re)placing the frozen version of
           this POD field.'''
        # Ensure a file from the correct type has been uploaded
        req = o.req
        format = req.podFormat
        upload = req.uploadedFile
        if not upload or not upload.name.endswith('.%s' % format):
            # A wrong file has been uploaded (or no file at all)
            msg = 'upload_invalid'
        else:
            # Store the uploaded file in the database
            self.freeze(o, req.template, format, secure=True, upload=upload)
            msg = 'action_done'
        return o.goto(o.referer, message=o.translate(msg))
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''An index for Date fields'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from DateTime import DateTime
from appy.database.indexes import Index

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class DateIndex(Index):
    '''Index for a Date field'''

    @classmethod
    def toIndexed(class_, value, field):
        '''Converts p_value, a DateTime instance, to its internal representation
           in the index.'''
        if not value: return
        if isinstance(value, int):
            # p_value is already an internal index representation, ready to be
            # indexed (probably a default value to index from
            # p_field.emptyIndexValue).
            return value
        # p_value is a DateTime instance that we will convert to an integer
        # ~~~
        # This code is inspired by the Zope catalog
        year,month,day,hours,minutes,seconds,zone = value.toZone('UTC').parts()
        r = (((year * 12 + month) * 31 + day) * 24 + hours) * 60 + minutes
        # Flatten to precision
        if field.indexPrecision > 1:
            r -= r % field.indexPrecision
        return r

    @classmethod
    def toTerm(class_, value, field):
        '''The passed p_value may already be an integer value (=the internal
           representation for a date in the index).'''
        return value if isinstance(value, int) \
                     else class_.toIndexed(value, field)

    @classmethod
    def fromIndexed(class_, value, field):
        '''Converts p_value, the internal integer representation of a date,
           into a DateTime instance.'''
        # p_indexed represents a number of minutes
        minutes = value % 60
        value = (value - minutes) / 60 # The remaining part, in hours
        # Get hours
        hours = value % 24
        value = (value - hours) / 24 # The remaining part, in days
        # Get days
        day = value % 31
        if day == 0: day = 31
        value = (value - day) / 31 # The remaining part, in months
        # Get months
        month = value % 12
        if month == 0: month = 12
        year = (value - month) / 12
        date= DateTime('%d/%d/%d %d:%d UTC'% (year, month, day, hours, minutes))
        return date.toZone(date.localZone())

    @classmethod
    def toString(class_, o, value):
        '''Returns the string representation of the date in p_value'''
        return o.tool.formatDate(value, format='%dt %d %mt %Y') if value \
                                                                else None
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



xhtmlInput = '''
<table width="500px">
 <tr>
  <td>Hello pixels</td>
  <td>Table properties</td>
 </tr>
</table>
<table style="width:50%">
 <tr>
  <td>Hello percentage</td>
  <td>Table properties</td>
 </tr>
</table>
<!-- We will ignore silly width=0px -->
<table width="0px">
  <tr>
    <td style="border:1px solid #000000">
      <p>1</p>
    </td>
  </tr>
</table>
'''



'''Functions for sending emails'''
import smtplib, socket, time
from email.MIMEMultipart import MIMEMultipart
from email.MIMEBase import MIMEBase
from email.MIMEText import MIMEText
from email.Utils import formatdate
from email import Encoders
from email.Header import Header
from appy.shared.utils import sequenceTypes

# ------------------------------------------------------------------------------
class MailConfig:
    '''Parameters for connecting to a SMTP server'''
    # Currently we don't check the connection to the SMTP server at startup
    testable = False

    def __init__(self, fromName=None, fromEmail='info@appy.org', replyTo=None,
                 server='localhost', port=25, secure=False,
                 login=None, password=None, enabled=True):
        # The name that will appear in the "from" part of the messages
        self.fromName = fromName
        # The mail address that will appear in the "from" part of the messages
        self.fromEmail = fromEmail
        # The optional "reply-to" mail address
        self.replyTo = replyTo
        # The SMTP server address and port
        if ':' in server:
            self.server, port = server.split(':')
            self.port = int(port)
        else:
            self.server = server
            self.port = int(port) # That way, people can specify an int or str
        # Secure connection to the SMTP server ?
        self.secure = secure
        # Optional credentials to the SMTP server
        self.login = login
        self.password = password
        # Is this server connection enabled ?
        self.enabled = enabled

    def init(self, tool): pass
    def getFrom(self):
        '''Gets the "from" part of the messages to send.'''
        if self.fromName: return '%s <%s>' % (self.fromName, self.fromEmail)
        return self.fromEmail

    def __repr__(self):
        '''Short string representation of this mail config, for logging and
           debugging purposes.'''
        r = '%s:%d' % (self.server, self.port)
        if self.login: r += ' (login as %s)' % self.login
        return r

# ------------------------------------------------------------------------------
def sendMail(config, to, subject, body, attachments=None, log=None,
             replyTo=None):
    '''Sends a mail, via the SMTP server defined in the p_config (an instance of
       appy.gen.mail.MailConfig above), to p_to (a single email recipient or a
       list of recipients). Every (string) recipient can be an email address or
       a string of the form "[name] <[email]>".

       p_attachments must be a list or tuple whose elements can have 2 forms:
         1. a tuple (fileName, fileContent): "fileName" is the name of the file
            as a string; "fileContent" is the file content, also as a string;
         2. a appy.fields.file.FileInfo instance.

       p_log can be a function accepting 2 args:
        - the message to log (as a string);
        - the second must be named "type" and will receive string
          "info", "warning" or "error".

       A p_replyTo mail address or recipient can be specified.
    '''
    start = time.time()
    if isinstance(to, str): to = [to]
    if not config:
        if log: log('Must send mail but no smtp server configured.')
        return
    # Just log things if mail is disabled
    fromAddress = config.getFrom()
    replyTo = replyTo or config.replyTo
    if not config.enabled or not config.server:
        if not config.server:
            msg = ' (no mailhost defined)'
        else:
            msg = ''
        if log:
            toLog = 'mail disabled%s: should send mail from %s to %d ' \
                    'recipient(s): %s.' % (msg, fromAddress, len(to),
                    ', '.join(to))
            if replyTo: toLog += ' (Reply to: %s).' % replyTo
            log(toLog)
            log('subject: %s' % subject)
            log('body: %s' % body)
        if attachments and log: log('%d attachment(s).' % len(attachments))
        return
    if log: log('sending mail from %s to %s (subject: %s).' % \
                (fromAddress, str(to), subject))
    # Create the base MIME message
    body = MIMEText(body, 'plain', 'utf-8')
    if attachments:
        msg = MIMEMultipart()
        msg.attach(body)
    else:
        msg = body
    # Add the header values
    msg['Subject'] = Header(subject, 'utf-8')
    msg['From'] = fromAddress
    msg['Date'] = formatdate(localtime=True)
    if replyTo: msg.add_header('reply-to', replyTo)
    if len(to) == 1:
        msg['To'] = to[0]
    else:
        msg['To'] = fromAddress
        msg['Bcc'] = ', '.join(to)
    # Add attachments
    if attachments:
        for attachment in attachments:
            # 2 possible forms for an attachment
            if isinstance(attachment, tuple) or isinstance(attachment, list):
                fileName, fileContent = attachment
            else:
                # a FileInfo instance
                fileName = attachment.uploadName
                f = file(attachment.fsPath, 'rb')
                fileContent = f.read()
                f.close()
            part = MIMEBase('application', 'octet-stream')
            part.set_payload(fileContent)
            Encoders.encode_base64(part)
            part.add_header('Content-Disposition',
                            'attachment; filename="%s"' % fileName)
            msg.attach(part)
    # Send the email
    try:
        smtpServer = smtplib.SMTP(config.server, port=config.port)
        if config.secure:
            # The next line is only required for Python > 2.6, that calls it
            # within method "starttls" below.
            smtpServer.ehlo()
            smtpServer.starttls()
            # The next line is only required for Python < 2.5: more recent
            # Pythons call it again after the TLS connection has been
            # established, at the end of method "starttls".
            smtpServer.ehlo()
        if config.login:
            smtpServer.login(config.login, config.password)
        r = smtpServer.sendmail(fromAddress, to, msg.as_string())
        smtpServer.quit()
        if log:
            if r:
                log('could not send mail to some recipients. %s' % str(r),
                    type='warning')
            else:
                log('mail sent in %.2f secs.' % (time.time()-start))
    except smtplib.SMTPException, e:
        if log:
            log('%s: mail sending failed (%s)' % (config, str(e)), type='error')
    except socket.error, se:
        if log:
            log('%s: mail sending failed (%s)' % (config, str(se)),type='error')
# ------------------------------------------------------------------------------



'''This file is responsible for perfoming the backup of a Appy site.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import time, shutil
from pathlib import Path

from DateTime import DateTime

from appy.pod import formatsByTemplate
from appy.model.utils import Object as O
from appy.utils import executeCommand, Traceback
from appy.utils.path import getShownSize, getOsTempFolder

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# If you want to configure the backup of your Appy server, place, in the main
# config's "backup" attribute, an instance of the Config class as defined below.

# Moreover, you'll have to define a job, in your main config's "jobs" attribute,
# defined with the tool method named "backup", in order to enable the backup, at
# the desired frequency and planning. For more information about jobs, check
# appy/server/scheduler.py.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
PATH_KO    = 'Backup config: folder "%s" does not exist. Please create it.'
BKP_DIS    = 'Backup cannot take place unless you configure it in ' \
             'config.backup.'
BKP_TYPE   = 'Backup type: %s (full is @%s).'
CMD        = 'Running: %s...'
DONE       = "Done in %.2f''."
S_SIZE     = 'Source file size is %s.'
PACKING    = 'Packing %s (%s)...'
COPY_LOG   = 'Copy %s to %s.'
DETAILS    = "\nMore details to be found in site's app.log file."
DELETED    = 'Deleted from %s: %s'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Backup configuration'''

    def __init__(self):
        # The (absolute) folder where to store backup files. It must exist. Appy
        # will create 3 sub-folders in it:
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # zodb  | will store copies of the appy.fs file (or named differently if
        #       | you have changed the default naming convention). 2 copies will
        #       | be created: backup1.fs and backup2.fs. Everytime the backup is
        #       | launched, it will copy appy.fs to the less recent version
        #       | among these 2 files. That way, if the copy crashes, there is
        #       | always an intact copy in the "zodb" folder.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # files | will store a copy of the DB-controlled file system
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # logs  | will create a backup of log files app.log and site.log
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # If attribute "path" is None, the backup is considered to be disabled.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        self.path = None

        # Just before and after having copied data and logs to p_self.path, 2
        # commands may be executed. This is typically used to mount/unmount
        # p_self.path if it located on a network drive. Please specify each
        # command as a list: the first element being the name of the command,
        # the subsequent ones being args.
        self.beforeCommand = None
        # Example: ['mount', '-t', 'cifs', '//some/server', '/mnt/backup']
        self.afterCommand = None
        # Example: ['umount', '/mnt/backup']

        # The list of email recipients that will receive an email containing
        # details about the backup procedure.
        self.emails = []

        # The name of the day corresponding to a full backup, as a string that
        # can be: Mon, Tue, Wed, Thu, Fri, Sat, Sun or All. A full backup
        # distinguishes itself from a standard backup by adding:
        # - a database "pack". The ZODB stores any change made to data.
        #   "Packing" it removes previous versions of objects for which a more
        #   recent version is found. It can substantially reduce the database
        #   size.
        # - a backup of the log files. app.log and site.log are copied to the
        #   backup folder and emptied.

        # On Appy sites with a high lever of write operations, the database may
        # grow in size very quickly. In that case, it may be appropriate to
        # perform a "full" backup everyday, by using value "All".
        self.fullDay = 'Sun'

        # The following attributes stores the name of a method on the tool that
        # the backup will call after the backup procedure has been executed. You
        # could also directly define a job with this method, but by specifying
        # it here, it allows you to define a unique "backup" job, that will
        # perform the backup + run your nightlife actions.

        # Pay attention to this method: if it raises an error, the whole backup
        # job will be canceled, in order to avoid committing partial results.
        self.method = None

        # Files produced by LibreOffice in server mode, asked by the Appy
        # server, may still be in the OS temp folder. Files whose extensions are
        # listed here will be removed from the OS temp folder.
        exts = ()
        for formats in formatsByTemplate.values():
            exts += formats
        self.tempExtensions = exts

        # Must the Appy server restart at the end of the job ? If you want to
        # enable this, set an integer value representing the number of seconds
        # to wait before restarting the server. This is required to let the
        # server terminate his ongoing transaction.
        self.restart = 10

    def check(self):
        '''Checks the backup configuration'''
        # If no backup folder is defined, the backup is supposed to be disabled
        path = self.path
        if not path: return
        # Check if the backup folder exists
        path = Path(path)
        if not path.is_dir():
            raise Exception(PATH_KO % self.path)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Backup:
    '''Performs backups'''

    # Options to rsync when copying DB-controlled files
    filesOptions = ['rsync', '-Paz', '--no-o', '--no-g', '--no-l', '--delete',
                    '--quiet', '--exclude', '*.fs*', '--exclude', '*.log']

    # Type of backup: full or not
    type = { True: 'full', False: 'standard' }

    def __init__(self, tool):
        self.tool = tool
        self.config = tool.config.backup
        self.server = tool.H().server
        # All log messages will also be stored in this list, to be sent by email
        self.messages = []

    def isFull(self):
        '''Returns True if the backup must be a full backup'''
        day = self.config.fullDay
        return (day == 'All') or time.asctime().startswith(day)

    def log(self, message):
        '''Logs this m_message and adds it to p_self.messages'''
        self.tool.log(message)
        self.messages.append(message)

    def logDone(self, start):
        '''Logs the "done" message, with the time elapsed since p_start'''
        self.log(DONE % (time.time() - start))

    def runCommand(self, args, text=None):
        '''Runs and logs a commmand made of p_args'''
        start = time.time()
        self.log(CMD % ' '.join(args))
        if text: self.log(text)
        # Run the sync
        out, err = executeCommand(args)
        if err: self.log(err)
        elif out: self.log(out)
        else: self.logDone(start)

    def runMethod(self):
        '''Runs the custom method as defined in the backup config. Returns True
           if the method did not raise an exception.'''
        try:
            method = self.config.method
            start = time.time()
            self.log(CMD % ('tool::%s' % method))
            getattr(self.tool, method)()
            self.logDone(start)
            return True
        except Exception as err:
            self.log(Traceback.get().strip(), type='error')

    def copyDatabase(self):
        '''Copy the .fs file to the <backup folder>/zodb, using rsync'''
        # Create the destination folder if it does not exist
        destFolder = self.config.path / 'zodb'
        if not destFolder.is_dir(): destFolder.mkdir()
        # Get the last modification dates for backup files
        modified1 = modified2 = None
        for backupFile in destFolder.glob('backup*.fs'):
            modified = DateTime(backupFile.stat().st_mtime)
            if backupFile.name == 'backup1.fs':
                modified1 = modified
            elif backupFile.name == 'backup2.fs':
                modified2 = modified
        # Select the file whose modification date is the less recent one
        if modified1 and modified2:
            nb = 1 if modified1 < modified2 else 2
        else:
            nb = 1 if not modified1 else 2
        # Prepare the rsync command
        sourceFile = self.tool.config.database.filePath
        fileSize = getShownSize(sourceFile.stat().st_size)
        args = ['rsync', '--inplace', '--no-whole-file',
                str(sourceFile), str(destFolder / ('backup%d.fs' % nb))]
        # Run the sync
        self.runCommand(args, text=S_SIZE % fileSize)

    def copyFiles(self):
        '''Copy the files from the DB-controlled file system the
           <backup folder>/files, using rsync.'''
        # Create the destination folder if it does not exist
        destFolder = self.config.path / 'files'
        if not destFolder.is_dir(): destFolder.mkdir()
        # Prepare the rsync command
        args = list(self.filesOptions)
        args.append('%s/' % self.tool.config.database.binariesFolder)
        args.append(str(destFolder))
        # Run the sync
        self.runCommand(args)

    def copyLogs(self):
        '''Copy the log files to <backup folder>/logs and empty them'''
        # Create the destination folder if it does not exist
        destFolder = self.config.path / 'logs'
        if not destFolder.is_dir(): destFolder.mkdir()
        # Determine the "date" part to encrust in the name of the backup files
        prefix = DateTime().strftime('%Y_%m_%d_%H_%M')
        # Walk log files
        config = self.tool.config.log
        for logFile in (config.site.path, config.app.path):
            # Determine the name of the target
            destName = '%s.%s.log' % (logFile.stem, prefix)
            destPath = str(destFolder / destName)
            # Launch the copy
            self.log(COPY_LOG % (logFile.name, destPath))
            shutil.copyfile(str(logFile), destPath)
            # Clean the content of the log file
            with open(logFile, 'w'): pass

    def sendMails(self):
        '''Send details about the operation, as stored in p_self.messages, to
           the mail recipients as defined in the backup config.'''
        tool = self.tool
        # Determine a subject for the mail
        subject = 'Appy / %s / Nightlife > %s' % \
                  (tool.config.model.appName, self.config.path)
        # The body is made of messages collected in p_self.messages
        body = '\n'.join(self.messages)
        # Send the mail
        tool.sendMail(self.config.emails, subject, body)

    def cleanTempFiles(self):
        '''Cleans temp files possibly produced by LibreOffice and left in the OS
           temp folder.'''
        # Do nothing if no file extension is configured
        exts = self.config.tempExtensions
        if not exts: return
        # Browse the OS temp folder, looking for files to delete
        tempFolder = Path(getOsTempFolder())
        counts = O()
        for ext in exts:
            count = 0
            for fileName in tempFolder.glob('*.%s' % ext):
                fileName.unlink()
                count += 1
            counts[ext] = count
        # Log deletions
        deleted = ['%s: %d' % (ext, count) for ext, count in counts.d().items()]
        self.log(DELETED % (tempFolder, ' - '.join(deleted)))

    def run(self):
        '''Performs the backup and related tasks'''
        config = self.config
        tool = self.tool
        handler = tool.H()
        # The backup can't be done unless correcty configured
        if not config or not config.path:
            self.log(BKP_DIS, type='warning')
            return
        # Is that a full backup or not ?
        full = self.isFull()
        self.log(BKP_TYPE % (self.type[full], config.fullDay))
        # Execute config.method if defined
        if config.method:
            success = self.runMethod()
            # Stop here if the method produced an error, in order to avoid
            # committing some partial result.
            if not success:
                handler.commit = False
                return
        # config.method may have set handler.commit to False. Force it to True,
        # because in the context of the currently running "backup" job, a commit
        # is always required (a minima, temp objects are removed from the
        # database).
        handler.commit = True
        # Remove temp objects within the database
        database = self.server.database
        logger = self.server.loggers.app
        database.cleanTemp(handler, logger)
        # Remove temp files in the OS temp folder
        self.cleanTempFiles()
        # Pack the database file if requested
        if full: database.pack(logger=logger)
        # Execute the "before command" if defined
        config.beforeCommand: self.runCommand(config.beforeCommand)
        # Copy the database itself (.fs file)
        self.copyDatabase()
        # Copy the files from the DB-controlled file system
        self.copyFiles()
        # Copy the log files when relevant
        if full: self.copyLogs()
        # Execute the "after command" if defined
        config.afterCommand: self.runCommand(config.afterCommand)
        # Send operation details by email when relevant
        self.messages.append(DETAILS)
        if config.emails: self.sendMails()
        # Restart the Appy server if required
        if config.restart: tool.restart(wait=config.restart)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''Appy allows you to create easily complete applications in Python'''

# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.
# ------------------------------------------------------------------------------
import os.path
import version
commercial = False

# ------------------------------------------------------------------------------
def getPath(): return os.path.dirname(__file__)

def getVersion():
    '''Returns a string containing the short and verbose Appy version'''
    if version.short == 'dev': return 'dev'
    return '%s (%s)' % version.short, version.verbose

def versionIsGreaterThanOrEquals(v):
    '''This method returns True if the current Appy version is greater than or
       equals p_v. p_v must have a format like "0.5.0".'''
    if version.short == 'dev':
        # We suppose that a developer knows what he is doing, so we return True
        return True
    else:
        paramVersion = [int(i) for i in v.split('.')]
        currentVersion = [int(i) for i in version.short.split('.')]
        return currentVersion >= paramVersion

# ------------------------------------------------------------------------------
class Object:
    '''At every place we need an object, but without any requirement on its
       class (methods, attributes,...) we will use this minimalist class.'''

    def __init__(self, **fields):
        for k, v in fields.iteritems():
            setattr(self, k, v)

    def __repr__(self):
        res = u'<O '
        for attrName, attrValue in self.__dict__.iteritems():
            v = attrValue
            if hasattr(v, '__repr__'):
                v = v.__repr__()
            try:
                res += u'%s=%s ' % (attrName, v)
            except UnicodeDecodeError:
                res += u'%s=<encoding problem> ' % attrName
        res  = res.strip() + '>'
        return res.encode('utf-8')

    def __nonzero__(self): return bool(self.__dict__)

    def d(self): return self.__dict__

    def get(self, name, default=None): return getattr(self, name, default)

    def __getitem__(self, k):
        '''Dict-like attribute get'''
        return getattr(self, k)

    def __setitem__(self, k, v):
        '''Dict-like attribute set'''
        self.__dict__[k] = v

    def __eq__(self, other):
        '''Equality between objects is, like standard Python dicts, based on
           equality of all their attributes and values.'''
        if isinstance(other, Object):
            return self.__dict__ == other.__dict__
        return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def update(self, other):
        '''Includes information from p_other into p_self'''
        for k, v in other.__dict__.iteritems():
            setattr(self, k, v)

    def clone(self):
        r = Object()
        r.update(self)
        return r

# ------------------------------------------------------------------------------
class Hack:
    '''This class proposes methods for patching some existing code with
       alternative methods.'''
    @staticmethod
    def patch(method, replacement, klass=None):
        '''This method replaces m_method with a p_replacement method, but
           keeps p_method on its class under name
           "_base_<initial_method_name>_". In the patched method, one may use
           Hack.base to call the base method. If p_method is static, you must
           specify its class in p_klass.'''
        # Get the class on which the surgery will take place
        isStatic = klass
        klass = klass or method.im_class
        # On this class, store m_method under its "base" name
        name = isStatic and method.func_name or method.im_func.__name__
        baseName = '_base_%s_' % name
        if isStatic:
            # If "staticmethod" isn't called hereafter, the static functions
            # will be wrapped in methods.
            method = staticmethod(method)
            replacement = staticmethod(replacement)
        setattr(klass, baseName, method)
        setattr(klass, name, replacement)

    @staticmethod
    def base(method, klass=None):
        '''Allows to call the base (replaced) method. If p_method is static,
           you must specify its p_klass.'''
        isStatic = klass
        klass = klass or method.im_class
        name = isStatic and method.func_name or method.im_func.__name__
        return getattr(klass, '_base_%s_' % name)

    @staticmethod
    def inject(patchClass, klass, verbose=False):
        '''Injects any method or attribute from p_patchClass into klass.'''
        # As a preamble, inject methods and attributes from p_patchClass's base
        # classes, if any.
        for base in patchClass.__bases__:
            Hack.inject(base, klass, verbose=verbose)
        patched = []
        added = []
        # Inject p_patchClass' own methods and attributes
        for name, attr in patchClass.__dict__.items():
            if name.startswith('__'): continue # Ignore special methods
            # Unwrap functions from static methods
            try:
                className = attr.__class__.__name__
            except AttributeError:
                # In Python 2, classes themselves have no defined class
                className = 'type'
            if className == 'staticmethod':
                attr = attr.__get__(attr)
                static = True
            else:
                static = False
            # Is this name already defined on p_klass ?
            if hasattr(klass, name):
                hasAttr = True
                klassAttr = getattr(klass, name)
            else:
                hasAttr = False
                klassAttr = None
            if hasAttr and (className != 'type') and \
               callable(attr) and callable(klassAttr):
                # Patch this method via Hack.patch
                if static:
                    Hack.patch(klassAttr, attr, klass)
                else:
                    Hack.patch(klassAttr, attr)
                patched.append(name)
            else:
                # Simply replace the static attr or add the new static
                # attribute or method.
                setattr(klass, name, attr)
                added.append(name)
        if verbose:
            pName = patchClass.__name__
            cName = klass.__name__
            print '%d method(s) patched from %s to %s (%s)' % \
                  (len(patched), pName, cName, str(patched))
            print '%d method(s) and/or attribute(s) added from %s to %s (%s)'%\
                  (len(added), pName, cName, str(added))
# ------------------------------------------------------------------------------



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
from http import HTTPStatus
import inspect, pathlib, mimetypes, os.path, email.utils, collections

from DateTime import DateTime

import appy
from appy.utils.string import Normalize

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
MAP_VALUE_NOT_PATH = 'Values from the map must be pathlib.Path objects.'
PATH_NOT_FOUND     = 'Path "%s" was not found or is not a folder.'
RAM_ROOT_IN_MAP    = 'Ram root "%s" is also used as key for ' \
                     'appy.server.static.Config.map.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Config:
    '''Configuration options for static content served by the Appy HTTP
       server.'''
    def __init__(self, appPath, root='static'):
        # The root URL for all static content. Defaults to "static". Any static
        # content will be available behind URL starting with <host>:/<root>/...
        self.root = root
        # The following attribute identifies base URLs and their corresponding
        # folders on disk. For example, the entry with key "appy" maps any URL
        # <host>:/<root>/appy/<resource> to the actual <resource> on disk, at
        # /<some>/<path>/<resource>.
        appyPath = pathlib.Path(inspect.getfile(appy)).parent
        map = collections.OrderedDict()
        map['appy'] = appyPath / 'ui' / 'static'
        map[appPath.stem] = appPath / 'static'
        self.map = map
        # The above-mentioned attributes define how to map a URL to a resource
        # on disk. A second mechanism is available, mapping URLs to "RAM"
        # resources, directly loaded in memory as strings. Such RAM resources
        # will be stored in dict Static.ram defined on class Static below.
        # Attribute "ramRoot" hereafter defines the base path, after the
        # "static" part, allowing Appy to distinguish a RAM from a disk
        # resource. It defaults to "ram".
        self.ramRoot = 'ram'
        # Let's illustrate this mechanism with an example. Suppose that:
        # * ramRoot = "ram" ;
        # * the content of appy/ui/static/appy.css is loaded in a string
        #   variable named "appyCss";
        # * dict Static.ram = {'appy.css': appyCss}
        # The content of appy.css will be returned to the browser if request via
        #                  <host>/static/ram/appy.css
        # Beyond being probably a bit more performant than serving files from
        # the disk, this approach's great advantage it to be able to compute, at
        # server startup, the content of resources. The hereabove example was
        # not randomly chosen: Appy CSS files like appy.css are actually
        # template files containing variables that need to be replaced by their
        # actual values, coming from the app's (ui) config.
        # When adding keys in Static.ram, ensure the key is a filename-like
        # name: the MIME type will be deduced from the file extension.

        # Remember the date/time this instance has been created: it will be used
        # as last modification date for RAM resources.
        self.created = DateTime()

    def check(self):
        '''Checks that every entry in p_self.map is valid'''
        for key, path in self.map.items():
            # Paths must be pathlib.Path instances
            if not isinstance(path, pathlib.Path):
                raise Exception(MAP_VALUE_NOT_PATH)
            # Paths must exist and be folders
            if not path.is_dir():
                raise Exception(PATH_NOT_FOUND % path)
        # Ensure the RAM root is not used as key in self.map
        if self.ramRoot in self.map:
            raise Exception(RAM_ROOT_IN_MAP % self.ramRoot)

    def init(self, uiConfig):
        '''Reads all CSS files from all disk locations containing static content
           (in p_self.map) and loads them in Static.ram, after having replaced
           variables with their values from the app's (ui) config.'''
        for path in self.map.values():
            for cssFile in path.glob('*.css'):
                # Read the content of this file
                with cssFile.open('r') as f:
                    cssContent = f.read()
                    # Add it to Static.ram, with variables replaced
                    Static.ram[cssFile.name] = uiConfig.patchCss(cssContent)

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Static:
    '''Class responsible for serving static content'''

    # The dict of RAM resources (see doc in class Config hereabove)
    ram = collections.OrderedDict()

    @classmethod
    def notFound(class_, handler, config):
        '''Raise a HTTP 404 error if the resource defined by p_handler.parts was
           not found.'''
        path = '/%s/%s' % (config.root, '/'.join(handler.parts))
        code = HTTPStatus.NOT_FOUND # 404
        handler.log('app', 'error', '%d @ %s' % (code.value, path))
        resp = handler.resp
        resp.code = code
        resp.build()

    @classmethod
    def writeUnchanged(class_, handler):
        '''Returns the ad-hoc response if the file has not changed since the
           last time the browser has downloaded it.'''
        resp = handler.resp
        resp.code = HTTPStatus.NOT_MODIFIED # 304
        resp.build()

    @classmethod
    def write(class_, handler, path, modified, content=None, fileInfo=None,
              disposition='attachment', downloadName=None, enableCache=True):
        '''Serves, to the browser, the content of the file whose path on disk is
           given in p_path or whose content in RAM is given in p_content.'''
        # ~~~
        # If the file @ p_path has not changed since the last time the browser
        # asked it, return an empty response with code 304 "Not Modified". Else,
        # return file content with a code 200 "OK".
        # ~~~
        # If p_path corresponds to a DB-controlled file, his corresponding
        # p_fileInfo is given. In that case, p_disposition will be taken into
        # account. Else, it will be ignored, unless p_downloadName is specified.
        # ~~~
        # For privacy reasons, p_enableCache may be disabled (ie, for
        # potentially sensitive content from File fields). This way, it cannot
        # be stored in the browser cache.
        # ~~~
        browserDate = handler.headers.get('If-Modified-Since')
        modified = fileInfo.modified if fileInfo else modified
        smodified = email.utils.formatdate(modified, usegmt=True) # RFC 822
        if not browserDate or (smodified > browserDate):
            resp = handler.resp
            resp.code = HTTPStatus.OK
            # Identify MIME type
            set = resp.setHeader
            mimeType, encoding = mimetypes.guess_type(path)
            mimeType = mimeType or 'application/octet-stream'
            set('Content-Type', mimeType)
            # Define content disposition
            if fileInfo or downloadName:
                niceName = downloadName or fileInfo.uploadName
                disp = '%s;filename="%s"' % (disposition, niceName)
                set('Content-Disposition', disp)
            # ~~~ Manage caching ~~~
            if enableCache:
                set('Last-Modified', smodified)
                # Ensure there is no the cache-related header
                resp.removeHeader('Cache-Control')
                resp.removeHeader('Expires')
            else:
                set('Cache-Control', 'no-cache, no-store, must-revalidate')
                set('Expires', '0')
            # Write the file content to the socket
            path = None if content else path
            resp.build(content, path)
        else:
            class_.writeUnchanged(handler)

    @classmethod
    def writeFromDisk(class_, handler, path,
                      disposition='attachment', downloadName=None):
        '''Serve a static file from disk, whose path is p_path'''
        # The string version of p_path
        spath = str(path)
        class_.write(handler, spath, os.path.getmtime(spath),
                     disposition=disposition, downloadName=downloadName)

    @classmethod
    def writeFromRam(class_, handler, config):
        '''Serve a static file loaded in RAM, from dict Static.ram'''
        # p_handler.parts contains something starting with ['ram', ...]
        if len(handler.parts) == 1:
            class_.notFound(handler, config)
            return
        # Re-join the splitted path to produce the key allowing to get the file
        # content in Static.ram.
        key = '/'.join(handler.parts[1:])
        # Indeed, Static.ram is a simple (ordered) dict, not a hierachical dict
        # of dicts. Standard Appy resources (like appy.css) stored in Static.ram
        # have keys whose names are simple filename-like keys ("appy.css"). But
        # if you want to reproduce a complete "file hierarchy" in Static.ram by
        # adding path-like information in the key, you can do it. By computing
        # keys like we did hereabove, a URL like:
        #               <host>/static/ram/a/b/c/some.css
        # can be served by defining, in Static.ram, an entry with key
        #                      "a/b/c/some.css"
        content = class_.ram.get(key)
        if content is None:
            class_.notFound(handler, config)
            return
        class_.write(handler, key, config.created, content=content)

    @classmethod
    def get(class_, handler):
        '''Returns the content of the static file whose splitted path is defined
           in p_handler.parts.'''
        # Unwrap the static config
        config = handler.server.config.server.static
        # The currently walked path
        path = None
        # Walk parts
        for part in handler.parts:
            if path is None:
                # We are at the root of the search: "part" must correspond to
                # the RAM root or to a key from config.map.
                if part == config.ramRoot:
                    class_.writeFromRam(handler, config)
                    return
                elif part in config.map:
                    path = config.map[part]
                else:
                    class_.notFound(handler, config)
                    return
            else:
                path = path / part
                if not path.exists():
                    class_.notFound(handler, config)
                    return
        # We have walked the complete path: ensure it is a file
        if not path or not path.is_file():
            return class_.notFound(handler, config)
        # Read the file content and write it in the HTTP response
        class_.writeFromDisk(handler, path)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



'''PX stands for *P*ython *X*ML. It is a templating engine that reuses the pod
   engine to produce XML (including XHTML) from templates written as a mix of
   Python and XML.'''

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import xml.sax
from appy.model.utils import Object as O
from appy.pod.buffers import MemoryBuffer
from appy.xml import xmlPrologue, xhtmlPrologue
from appy.px.parser import PxParser, PxEnvironment

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
SPEC_KO      = 'wrong specifier. A specifier must be of the form ' \
               '"<objectName>*<pxName>", like in "tool*pxTest".'
SPEC_NAME_KO = 'name "%s" does not correspond to any object in the context.'
PX_NAME_KO   = 'attribute "%s.%s" does not exist or is not a Px instance.'

#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Px:
    '''Represents a (chunk of) PX code'''

    class Error(Exception): pass
    xmlPrologue = xmlPrologue
    xhtmlPrologue = xhtmlPrologue

    # There are various ways to define a PX
    STRING = 0 # PX code directly written down in a string
    FILE   = 1 # PX code read from a file
    METHOD = 2 # As a method that must return PX code

    def __init__(self, content, type=STRING, partial=True, template=None,
                 hook=None, prologue=None, css=None, js=None, name=None):
        '''Creates a PX object'''

        # ----------------------------------------------------------------------
        # If p_type is... | p_content is...
        # ----------------------------------------------------------------------
        #     STRING      | a chunk of PX code as a string 
        # ----------------------------------------------------------------------
        #      FILE       | the path to a file on disk, containing PX code
        # ----------------------------------------------------------------------
        #     METHOD      | (cheat mode) a method that will accept the PX
        #                 | context as unique arg and must return XML code, just
        #                 | as a PX would do. This allows to "implement" a PX
        #                 | with Python code, which may, in some cases, be
        #                 | appropriate for performance reasons.
        # ----------------------------------------------------------------------

        # If the PX code represents a complete XML file, p_partial is False.
        # Else, p_content will be surrounded by a root tag to be able to parse
        # it with a SAX parser.

        # If this PX is based on another PX template, specify the PX template in
        # p_template and the name of the p_hook where to insert this PX into the
        # template PX.

        # If a p_prologue is specified, it will be rendered at the start of the
        # PX result.

        # You can specify, in p_css and p_js, PX-specific CSS styles and
        # Javascript code, as strings. These strings will be surrounded by the
        # appropriate HTML tag ("style" for p_css and "script" for p_js) and
        # dumped just before the PX result. Note that if the PX is executed more
        # than once, its corresponding p_css and p_js will only be dumped before
        # the first PX result.

        # Get the PX content
        if type == Px.FILE:
            with open(content) as f: self.content = f.read()
        else:
            self.content = content
        self.isMethod = callable(content)
        # It this content a complete XML file, or just some part of it ?
        self.partial = partial
        # Is this PX based on a template PX ?
        self.template = template
        self.hook = hook
        # Is there some (XML, XHTML...) prologue to dump ?
        self.prologue = prologue
        
        # PX-specific CSS and JS code
        self.css = self.compact(css)
        self.js = self.compact(js)
        # A PX can have a name
        self.name = name or ''
        # A PX can be profiled (see m_profile below)
        self.profiler = None
        # Parse the PX
        if not self.isMethod:
            self.parse()

    def parse(self):
        '''Parses self.content and create the structure corresponding to this
           PX.'''
        if self.partial:
            # Surround the partial chunk with a root tag: it must be valid XML
            self.content = '<x>%s</x>' % self.content
        # Create a PX parser
        self.parser = PxParser(PxEnvironment(), self)
        # Parses self.content (a PX code in a string) with self.parser, to
        # produce a tree of memory buffers.
        try:
            self.parser.parse(self.content)
        except xml.sax.SAXParseException as spe:
            self.completeErrorMessage(spe)
            raise spe

    def compact(self, s):
        '''Removes single-line comments and unnecessary spaces in p_s,
           representing CSS or JS code.'''
        if not s: return
        r = []
        for line in s.split('\n'):
            line = line.strip()
            if not line.startswith('//'):
                r.append(line)
        return '\n'.join(r)

    def addCssJs(self, code, type):
        '''Adds this p_code, being CSS or JS, depending on p_type'''
        code = self.compact(code)
        existing = getattr(self, type)
        if existing:
            code = '%s\n%s' % (existing, code)
        setattr(self, type, code)

    def addCss(self, css): self.addCssJs(css, 'css')
    def addJs(self, js): self.addCssJs(js, 'js')

    def completeErrorMessage(self, parsingError):
        '''A p_parsingError occurred. Complete the error message with the
           erroneous line from self.content.'''
        # Split lines from self.content
        splitted = self.content.split('\n')
        i = parsingError.getLineNumber() - 1
        # Get the erroneous line, and add a subsequent line for indicating
        # the erroneous column.
        column = ' ' * (parsingError.getColumnNumber()-1) + '^'
        lines = [splitted[i], column]
        # Get the previous and next lines when present
        if i > 0: lines.insert(0, splitted[i-1])
        if i < len(splitted)-1: lines.append(splitted[i+1])
        parsingError._msg += '\n%s' % '\n'.join(lines)

    def __call__(self, context, applyTemplate=True, isTemplate=False):
        '''Renders the PX.

           If the PX is based on a template PX, we have 2 possibilities.
           1. p_applyTemplate is True. This case corresponds to the initial
              call to the current PX. In this case we call the template with a
              context containing, in the hook variable, the current PX.
           2. p_applyTemplate is False. In this case, we are currently executing
              the PX template, and, at the hook, we must include the current PX,
              as is, without re-applying the template (else, an infinite
              recursion would occur).
        '''
        # Developer, forget the following line
        if '_ctx_' not in context: context['_ctx_'] = context

        # Include, in the context, variable names for "reserved" PX chars
        if 'PIPE' not in context:
            context['PIPE'] = '|'
            context['SEMICOLON'] = ';'

        # The first PX to be called will be stored as key "_px_", while the
        # currently called (sub-)PX will be stored at key "_cpx_" ("c" stands
        # for "current").
        px = context.get('_px_')
        if not isTemplate:
            if px is None: # This is the first PX to be called
                context['_px_'] = self
            context['_cpx_'] = self

        # This PX call is probably one among a series of such calls, sharing the
        # same p_context. Within this context, we add a sub-dict at key "rt"
        # (*r*un-*t*ime) for counting the number of times every PX is called.
        # It allows us to include PX-specific CSS and JS code only once.
        if '_rt_' not in context: context['_rt_'] = {}

        if self.hook and applyTemplate:
            # Call the template PX, filling the hook with the current PX
            context[self.hook] = self
            r = self.template(context, isTemplate=True)
        else:
            # Start profiling when relevant
            profiler = self.profiler
            if profiler: profiler.enter(self.name)
            # Run the PX
            if self.isMethod:
                r = self.content(context)
            else:
                # Create a Memory buffer for storing the result
                env = self.parser.env
                result = MemoryBuffer(env, None)
                # Execute the PX
                env.ast.evaluate(result, context)
                # Get the PX result
                r = result.content
            # Count this call and include CSS and JS code when relevant
            pxId = id(self)
            rt = context['_rt_']
            if pxId in rt:
                rt[pxId] += 1
            else:
                rt[pxId] = 1
                # This is the first time we execute it: include CSS and JS code
                # if present.
                if self.js: r = ('<script>%s</script>\n' % self.js) + r
                css = self.css
                if css:
                    # Patch CSS if a CSS patcher is found
                    css = context['_css_'](css) if '_css_' in context else css
                    r = ('<style>%s</style>\n' % css) + r
            # Include the prologue
            if self.prologue:
                r = self.prologue + r
            # Stop profiling when relevant
            if profiler: profiler.leave()
        # Restore the previous PX in the context
        if px: context['_cpx_'] = px
        # Return the result
        return r

    def override(self, content, partial=True):
        '''Overrides the content of this PX with a new p_content (as a
           string).'''
        self.partial = partial
        self.content = content
        # Parse again, with new content
        self.parse()

    def profile(self, name, profiler):
        '''Enables profiling of this PX, that will be named p_name in the
           p_profiler's output.'''
        self.name = name
        self.profiler = profiler

    # Literals to convert via method p_convertValue below
    literals = {'True': True, 'False': False, 'None': None}

    @classmethod
    def convertValue(class_, value, tool):
        '''Converts p_value as found in the request to a value as can be added
           to a PX context.'''
        if isinstance(value, list):
            r = [class_.convertValue(v, tool) for v in value]
        elif not isinstance(value, str):
            r = value # Do not touch already "typed" values
        elif value.startswith(':'):
            # The value contains an object ID: convert it to a real object
            r = value[1:]
            if ':' not in r:
                r = tool.getObject(r)
            else:
                # A field name lies besides the object ID: create an initiator
                # object representing them.
                id, fname = r.split(':', 1)
                o = tool.getObject(id)
                r = tool.Initiator(tool, o.req, (o, o.getField(fname)))
        elif value in class_.literals:
            r = class_.literals[value]
        elif value.isdigit():
            r = int(value)
        else:
            r = value
        return r

    @classmethod
    def injectRequest(class_, context, req, tool, specialOnly=False):
        '''Inject all p_req(uest) values into the PX p_context (or only
           search-related entries if p_searchOnly is True).'''
        if not specialOnly:
            # Inject every (key, value) pair found in the request into the
            # context, with automatic type conversion for values.
            for name, value in req.items():
                setattr(context, name, class_.convertValue(value, tool))
        # When some special keys are found in the request, create related,
        # additional entries in the PX context. These keys correspond to base
        # Appy concepts that are regularly transmitted between the browser and
        # the server.
        if 'className' in req:
            # Add entry "class_"
            context.class_ = tool.model.classes.get(req.className)
        if 'search' in req:
            ui = tool.Search.get(req.search, tool, context.class_, context)
            context.uiSearch = ui
            context.resultModes = ui.getModes()
            context.mode = ui.getMode()

    @classmethod
    def createContext(class_, traversal, layout, Quote):
        '''Creates a context containing all base variables required to execute
           the PX named p_name.'''
        # Unwrap base objects from the p_traversal
        handler = traversal.handler
        guard = handler.guard
        tool = handler.tool
        req = handler.req
        o = traversal.o or tool
        # The logged user and the home object
        user = guard.user
        # The user language (2-letters ISO code)
        lang = guard.userLanguage
        # Is this language LTR or RTL ?
        dir, dleft, dright = handler.Languages.getDirection(lang)
        # The main app config
        config = handler.server.config
        # Define and return the context
        r = O(
          traversal=traversal, handler=handler, guard=guard, tool=tool, req=req,
          o=o, layout=layout, ui=tool.ui, _=o.translate, user=user,
          isAnon=user.isAnon(), Px=Px, config=config, field=traversal.field,
          appName=config.model.appName, q=Quote.js, url=tool.buildUrl,
          lang=lang, dir=dir, dleft=dleft, dright=dright,
          siteUrl=config.server.getUrl(handler), _css_=config.ui.patchCss)
        # Ensure keys "ajax" and "popup" are present
        for name in ('ajax', 'popup'):
            setattr(r, name, getattr(r, name) or False)
        # Compute and add the home object
        r.home = tool.computeHomeObject(user, r.popup)
        # Inject request values in it
        class_.injectRequest(r, req, tool)
        return r

    # Names of the base variables that must be present in the context of any
    # standard Appy PX.
    contextKeys = ('traversal', 'handler', 'guard', 'tool', 'req', 'o', 'home',
      'layout', 'popup', 'ajax', 'ui', '_', 'user', 'isAnon', 'Px', 'config',
      'appName', 'q', 'url', 'lang', 'dir', 'dleft', 'dright', '_css_')

    @classmethod
    def copyContext(class_, context, other):
        '''Copy base variables whose names are defined in p_class_.contextKeys
           from p_other to p_context.'''
        for key in class_.contextKeys:
            context[key] = other[key]

    # Names of the PXs being Appy home pages. Key indicates if "public" pages
    # are taken into account or not.
    appyHomePXs = {
      True:  {'home':None, 'homes':None, 'public':None},
      False: {'home':None, 'homes':None}
    }

    @classmethod
    def isHome(class_, ctx, orPublic=True):
        '''Returns True if the currently shown PX has the name of one of the
           Appy home PXs.'''
        return ctx._px_.name in class_.appyHomePXs[orPublic]

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Utility methods
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @staticmethod
    def truncateValue(value, width=20, suffix='...'):
        '''Truncates string p_value according to p_width'''
        width = width or 20 # p_width could be None
        if len(value) > width:
            value = value[:width] + suffix
        return value

    @staticmethod
    def truncateText(text, width=20, suffix='...', tag='abbr'):
        '''Truncates p_text to max p_width chars. If the text is longer than
           p_width, the truncated part is put in a "abbr" xhtml tag.'''
        width = width or 20 # p_width could be None
        if len(text) > width:
            text = '<%s title="%s">%s%s</%s>' % \
                   (tag, text, text[:width], suffix, tag)
        return text

    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    # Call a PX from a "specifier"
    #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

    @classmethod
    def callFromSpec(class_, specifier, o):
        '''Calls the PX mentioned in p_specifier, with the context defined on
           p_o.traversal.context.'''

        # A specifier is of the form <name>*<px>.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        # <name> | is a name that must be present in the PX context, pointing
        #        | to an object, module, class...
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
        #  <px>  | is the name of the PX instance that must exist as attribute
        #        | on the object denoted by <name>.
        #- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

        # Check if the p_specifier is valid
        parts = specifier.split('*', 1)
        if len(parts) != 2: raise class_.Error(SPEC_KO)
        # Does the named object exist in the context ?
        name, px = parts
        ctx = o.traversal.context
        if name not in ctx: raise class_.Error(SPEC_NAME_KO % name)
        # Does the named PX exist on the object ?
        obj = ctx[name]
        px = getattr(obj, px, None)
        if (px is None) or not isinstance(px, Px):
            raise class_.Error(PX_NAME_KO % (name, parts[1]))
        return px(ctx)
#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -



#- - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
# Copyright (C) 2007-2021 Gaetan Delannay

# This file is part of Appy.

# Appy is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# Appy is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE. See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# Appy. If not, see <http://www.gnu.org/licenses/>.

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
import random, hashlib, binascii

from appy.px import Px
from appy.model.fields import Field

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class PasswordGenerator:
    '''Class used to generate passwords'''

    # No "0" or "1" that could be interpreted as letters "O" or "l"
    passwordDigits = '23456789'

    # No letters i, l, o (nor lowercase nor uppercase) that could be misread
    passwordLetters = 'abcdefghjkmnpqrstuvwxyzABCDEFGHJKMNPQRSTUVWXYZ'

    @classmethod
    def get(k, minLength=8, maxLength=9):
        '''Generates and r_eturns a password whose length is between p_minLength
           and p_maxLength.'''
        # Compute the actual length of the challenge to encode
        length = random.randint(minLength, maxLength)
        r = ''
        for i in range(length):
            j = random.randint(0, 1)
            chars = (j == 0) and k.passwordDigits or k.passwordLetters
            # Choose a char
            r += chars[random.randint(0,len(chars)-1)]
        return r

#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
class Password(Field):
    '''Field allowing to edit and store a password'''

    view = cell = Px('''<x>******</x>''')
    edit = Px('''
     <x for="confirm in (False, True)"
        var2="placeholder=field.getPlaceholder(o, confirm);
              inputId=confirm and ('%s_confirm' % name) or name">
      <input type="password" id=":inputId" name=":inputId"
             size=":field.getInputSize()" style=":field.getInputSize(False)"
             maxlength=":field.maxChars" placeholder=":placeholder"/>
      <br if="not confirm"/>
     </x>''')

    def __init__(self, validator=None, multiplicity=(0,1), show=True,
      page='main', group=None, layouts=None, move=0, readPermission='read',
      writePermission='write', width='25em', height=None, maxChars=None,
      colspan=1, master=None, masterValue=None, focus=False, historized=False,
      mapping=None, generateLabel=None, label=None, sdefault='', scolspan=1,
      swidth=None, sheight=None, persist=True, placeholder=None, view=None,
      cell=None, edit=None, xml=None, translations=None, minLength=8):
        # The minimum length for this password
        self.minLength = minLength
        # Call the base constructor
        Field.__init__(self, validator, multiplicity, None, None, show, page,
          group, layouts, move, False, True, None, None, False, None,
          readPermission, writePermission, width, height, maxChars, colspan,
          master, masterValue, focus, historized, mapping, generateLabel, label,
          sdefault, scolspan, swidth, sheight, persist, False, view, cell, edit,
          xml, translations)
        # A potential placeholder (see homonym attribute in string.py)
        self.placeholder = placeholder

    def getPlaceholder(self, o, confirm):
        '''Returns a placeholder for the field if defined'''
        if confirm:
            # Set a specific label as placeholder for the input field allowing
            # to confirm the password.
            r = o.translate('password_confirm')
        else:
            # Define the placeholder for the base field
            r = self.getAttribute(o, 'placeholder') or ''
            # Use the field label if a placeholder must be set but no label is
            # explicitly defined.
            r = o.translate(self.labelId) if r is True else r
        return r

    def validateValue(self, o, password):
        '''Is p_password valid ?'''
        # Password must have a minimum length
        if len(password) < self.minLength:
            return o.translate('password_too_short', mapping={'nb': 8})
        # Ensure the "confirm" value is filled and is the same as p_password
        if o.req['%s_confirm' % self.name] != password:
            return o.translate('passwords_mismatch')

    def encrypt(self, password, prefix=True, salt=None):
        '''Encrypt clear p_password with the SSHA scheme. If p_prefix is True,
           the password is prefixed with the SSHA scheme ("${SSHA}"). If p_salt
           is not given, it will be computed.'''
        if salt is None:
            # Generate a salt made of 7 chars
            salt = ''
            for n in range(7):
                salt += chr(random.randrange(256))
            salt = salt.encode()
        # Use SHA-1 algorithm to encrypt the password
        r = hashlib.sha1(password.encode() + salt).digest() + salt
        # Base64-encode the result
        r = binascii.b2a_base64(r)[:-1]
        # Prefix the result with the SSHA prefix when appropriate
        if prefix:
            r = b'{SSHA}' + r
        return r

    def check(self, o, password):
        '''Return True if the clear p_password corresponds the password as
           encrypted for this field on p_o.'''
        # Get the encrypted password
        encrypted = o.values.get(self.name)
        if not encrypted: return
        # Remove the scheme prefix
        encrypted = encrypted[6:]
        # Base64-decode it
        try:
            base = binascii.a2b_base64(encrypted)
        except binascii.Error:
            # Not valid base64
            return
        # Encrypt p_password and compare it with the encrypted password
        return self.encrypt(password, prefix=False, salt=base[20:]) == encrypted

    def generate(self, maxLength=9):
        '''Generate a password of at most m_maxLength chars'''
        return PasswordGenerator.get(self.minLength, maxLength)

    def set(self, o, password=None, log=True, maxLength=9):
        '''Sets a p_password on p_o for this password field. If p_password is
           not given, a password will be generated, made of at most p_maxLength
           chars. This method returns the generated password (or simply
           p_password if no generation occurred).'''
        if password is None:
            # Generate one
            password = self.generate(maxLength)
            msgPart = 'generated'
        else:
            msgPart = 'changed'
        self.store(o, password)
        # Log the operation when requested
        if log: self.log('password %s for %s.' % (msgPart, login))
        return password

    def store(self, o, value):
        '''Encrypts the clear password given in p_value'''
        if not self.persist: return
        # If p_value is None, store it as is
        o.values[self.name] = None if value is None else self.encrypt(value)
#  - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
